# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/inference/infer_gate_hist.ipynb.

# %% auto 0
__all__ = ['get_tensor_gate_length', 'get_circuit_gate_length']

# %% ../../src/inference/infer_gate_hist.ipynb 2
from ..imports import *

# %% ../../src/inference/infer_gate_hist.ipynb 4
def get_tensor_gate_length(clr_tensor, padding_token=0):
    '''Careful with padding tokens!'''
    assert clr_tensor.dim() == 3 #[b, s, t]
    
    collabsed_clr_tensor = (clr_tensor != padding_token).to(torch.int8)
    red_clr_tensor = torch.sum(collabsed_clr_tensor, dim=1)  # [b, t]
    return torch.count_nonzero(red_clr_tensor, dim=1)    # [b]

# %% ../../src/inference/infer_gate_hist.ipynb 5
def get_circuit_gate_length(qcs): 
    lengths = torch.zeros(len(qcs), dtype=int)    
    for i,qc in enumerate(qcs):         
        if hasattr(qc, "data"):   
            lengths[i] = len(qc.data) 
    return lengths



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/config_model.ipynb.

# %% auto 0
__all__ = ['Config_Model']

# %% ../../src/models/config_model.ipynb 3
from ..imports import *
from ..config_loader import *
from ..util import *
from datetime import datetime

# %% ../../src/models/config_model.ipynb 5
class Config_Model(nn.Module):
    """A basic `nn.Module` with IO functionality."""
    def __init__(self): super().__init__()
    
    #---------------------
    
    def get_config(self, save_path=None, without_metadata=False):
        if not without_metadata:       
            config = {}
            config["target"]         = class_to_str(type(self)) 
            config["save_path"]      = save_path
            config["save_datetime"]  = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
            config["params"]         = self.params_config  
        else:
            config = self.params_config  
        
        self.config = config        
        return config
    
    def store_model(self, config_path: str=None, save_path: str=None, without_metadata=False):        
    
        config = self.get_config(save_path, without_metadata)
    
        if exists(config_path):
            if without_metadata: save_dataclass_yaml(config, config_path)
            else               : save_dict_yaml(config, config_path)            
                       
        if exists(save_path):
            torch.save(self.state_dict(), save_path)     
    
    #---------------------
    
    @staticmethod
    def from_config(config, device: torch.device, save_path: str=None):  
        """Use this if we have a loaded config. Maybe within other classes (e.g. pipeline and nested models)"""
        
        model = instantiate_from_config(config)
        model = model.to(device) 
        print(f"[INFO]: `{class_to_str(type(model))}` instantiated from given config on {device}.")
        
        #--------------------------------        
        if not exists(save_path):            
            if "save_path" in config:
                save_path = config["save_path"]
            else:
                print("[INFO]: Found no key `save_path` path in config.")
                                  
        if exists(save_path):
            model.load_state_dict(torch.load(save_path, map_location=torch.device(device).type, weights_only=True), strict=True)
        else:
            print(f"[INFO]: `{class_to_str(type(model))}`. No save_path` provided. No state dict loaded.")

        return model
    
    @staticmethod
    def from_config_file(config_path, device: torch.device, save_path: str=None):    
        config = load_config(config_path)
        return Config_Model.from_config(config, device, save_path)       



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/scheduler/scheduler.ipynb.

# %% auto 0
__all__ = ['Scheduler']

# %% ../../src/scheduler/scheduler.ipynb 3
from ..imports import *
from ..util import virtual
from ..config_loader import *

# %% ../../src/scheduler/scheduler.ipynb 4
class Scheduler:
    """Base class for all diffusion schedulers"""
    def __init__(self): 
        pass
    
    @virtual
    def set_timesteps(self): pass
     
    @virtual
    def step(self): pass

    @virtual
    def add_noise(self): pass

    #---------------------------------------

    def get_config(self, without_metadata=False):
        if not without_metadata:       
            config = {}
            config["target"] = class_to_str(type(self))
            config["params"] = self.params_config  
        else:
            config = self.params_config  

        self.config = config        
        return config
    
    @property
    @virtual
    def params_config(self): return None

    #---------------------------------------

    def unsqueeze_vector_to_shape(self, vec, shape):
        vec = vec.flatten()
        while len(vec.shape) < len(shape):
            vec = vec.unsqueeze(-1)
        return vec 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/dataset/dataset_helper.ipynb.

# %% auto 0
__all__ = ['check_duplicate_in_dataset', 'check_duplicates_in_dataset_python', 'check_duplicates_in_dataset',
           'shuffle_tensor_dataset', 'get_unique_elements_indices', 'uniquify_tensor_dataset', 'balance_tensor_dataset',
           'map_old_tensor_to_new']

# %% ../../src/dataset/dataset_helper.ipynb 3
from ..imports import *
from ..config_loader import *

# %% ../../src/dataset/dataset_helper.ipynb 5
def check_duplicate_in_dataset(x, dataset):
    """Check if 'x' is in 'dataset'"""
    # x        ...  [   *]
    # dataset  ...  [n, *]
    
    comp = (dataset==x)  
    comp = torch.reshape(comp, [comp.shape[0], -1])  
    comp = torch.all(comp, dim=1)
    
    num  = comp.nonzero().squeeze().numel()  
    return bool(num)

# %% ../../src/dataset/dataset_helper.ipynb 6
def check_duplicates_in_dataset_python(xs, dataset):
    cnt = 0
     
    raise NotImplementedError("")
    
    # f = lambda x: int(check_duplicate_in_dataset(x, dataset))
    # res = async_loop_consumer(f, xs)
    # cnt = sum(res)
    
    comp = []
    
    for i,x in enumerate(xs):
        if check_duplicate_in_dataset(x, dataset):
            #print(f"[INFO] Duplicate in dataset at index={i}")
            comp.append(i)
            cnt += 1           
    # print(f"[INFO] Found {cnt}/{xs.shape[0]} duplicates in dataset of {dataset.shape[0]}.")
    
    return cnt, comp

# %% ../../src/dataset/dataset_helper.ipynb 7
def check_duplicates_in_dataset(xs, dataset, return_ind=False, invert=False):
    '''
    Checks if `xs` is are `dataset`. Boolean `invert` changes if we count duplicates (False) or ones that are not in dataset (True). 
    Uses `torch.vmap` which copies `dataset` for every element in `xs`.
    '''
    
    # x        ...  [c, *]
    # dataset  ...  [n, *]
        
    def get_comp(x, dataset):
        comp = (dataset==x) 
        comp = torch.reshape(comp, [comp.shape[0], -1])  
        comp = torch.all(comp, dim=1)  
        return comp 
            
    get_comp_map = torch.vmap(get_comp, in_dims=(0, None)) 
    
    # try: #could run out of memory bcs dataset is copied each time
    
    comp = get_comp_map(xs, dataset)
    
    if invert: comp = torch.all(comp==False, dim=1) 
    else:      comp = torch.any(comp, dim=1)
    
    comp = comp.nonzero()
    num  = comp.shape[0]                           
    
    # except Exception as er:
    #     print("[WARNING] check_duplicates_in_dataset:", er)
    #     print("We will use python instead.")
    #     raise NotImplementedError("")
    #     # cnt, comp = check_duplicates_in_dataset_python(xs, dataset)
        
    if return_ind: return num, comp.squeeze()  #comp is [i_xs, i_dataset] pairs
    return num

# %% ../../src/dataset/dataset_helper.ipynb 11
def shuffle_tensor_dataset(x, y=None, *z):
    '''Assumes numpy or tensor objects with same length.'''
    rand_indx = torch.randperm(x.shape[0])
    
    if exists(y):
        assert x.shape[0] == y.shape[0]       
        for iz in z: assert x.shape[0] == iz.shape[0]         
        return x[rand_indx], y[rand_indx], *(iz[rand_indx] for iz in z)
    
    return x[rand_indx]

# %% ../../src/dataset/dataset_helper.ipynb 12
def get_unique_elements_indices(tensor):
    '''Returns indices of unique_elements in `tensor`.'''
    tensor_unique, ptrs, cnt = torch.unique(tensor, dim=0, return_inverse=True, return_counts=True)
    _, ind_sorted = torch.sort(ptrs, stable=True) #e.g. gets the index that points to zero at pos [0]
    
    cum_sum = cnt.cumsum(0)
    cum_sum = torch.cat((torch.tensor([0], device=tensor.device), cum_sum[:-1]))
        
    return tensor_unique, ind_sorted[cum_sum]

# %% ../../src/dataset/dataset_helper.ipynb 13
def uniquify_tensor_dataset(x, y=None, *z):
    '''`x` has to be tensor, assumes numpy or tensor obj for `y` and `z`'''
    x, x_idx = get_unique_elements_indices(x)
    x_idx    = x_idx.cpu()
    
    if exists(y):
        assert x.shape[0] == y[x_idx].shape[0]   
        for iz in z: assert x.shape[0] == iz[x_idx].shape[0]
        return x, y[x_idx], *(iz[x_idx] for iz in z)
    
    return x

# %% ../../src/dataset/dataset_helper.ipynb 14
def balance_tensor_dataset(x, y, *z, samples: int=None, make_unique: bool=True, y_uniques=None, shuffle_lables: bool=True, add_balance_fn: callable=None):
    '''Assumes `x` is tensor and `y` is tensor or numpy.'''
    
    y_type = type(y)
    assert y_type in [np.ndarray, torch.Tensor]
    
    #------------------------------
    
    if make_unique:
        x, y, *z = uniquify_tensor_dataset(x, y, *z)
        assert x.shape[0] == y.shape[0]
        
        #bcs unique sorts, we need to shuffle the dataset before picking the first 'samples' entries
        x, y, *z = shuffle_tensor_dataset(x, y, *z)      
         
    #------------------------------
    
    if y_type == np.ndarray: y_uniques_temp, y_uniques_cnt =    np.unique(y, return_counts=True, axis=0)
    else:                    y_uniques_temp, y_uniques_cnt = torch.unique(y, return_counts=True, dim=0)
    
    if y_uniques is None: y_uniques = y_uniques_temp
    if   samples is None:    
        if y_type == np.ndarray: samples =    np.min(y_uniques_cnt)   # the actual balancing count
        else:                    samples = torch.min(y_uniques_cnt)
         
    ind = list()   
    for y_unique in y_uniques:
        
        if y_type == np.ndarray:
            comp    = (y==y_unique)
            indices = np.squeeze(np.nonzero(comp))
            indices = indices if indices.ndim > 0 else indices[None]
            
        else:
            comp    = torch.all(y==y_unique, dim=1)
            indices = comp.nonzero().squeeze().cpu()
            indices = indices if indices.dim() > 0 else indices[None]
        
        #special add balncing, e.g., for circuit length
        if add_balance_fn is not None: indices = add_balance_fn(indices, x, y, *z)
            
        ind.append(indices[:samples]) #limit samples
        
    if y_type == np.ndarray: ind = np.concatenate(ind, axis=0)
    else:                    ind = torch.cat(ind, dim=0)
    
    xb = x[ind]
    yb = y[ind]
    zb = (iz[ind] for iz in z)
        
    #shuffle again bcs we are sorted by label; normaly we don't care and we shuffle with dataset obj anyways, but needed if valid split afterwards
    if shuffle_lables: xb, yb, *zb = shuffle_tensor_dataset(xb, yb, *zb) 
        
    return xb, yb, *zb

# %% ../../src/dataset/dataset_helper.ipynb 16
def map_old_tensor_to_new(x):
    raise DeprecationWarning("[WARNING] There really should be no more old tensors arround .... delete them")
    print("[WARNING] There really should be no more old tensors arround .... delete them")
    
    b, gc, bits, t = x.shape
    
    x = x.reshape((b, gc//3, 3, bits, t))      # [b, g-c,  bits, t] -> [b, g, c, bits, t]   
    x = torch.argmax(x, dim=2)                 # [b, g, c, bits, t]-> [b, g, bits, t]   
    
    gate = torch.concat([torch.zeros_like(x[:,:1]), x], dim=1)  # add zeros for empty token   
    gate = torch.argmax(gate, dim=1)
    
    control_target = torch.sum(x, dim=1)
    mapped_tensor  = torch.zeros_like(control_target)
    mapped_tensor[control_target==1] = -1
    mapped_tensor[control_target==2] =  1
     
    return gate * mapped_tensor   # is now [b, space, time] with elements +-gate_number



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/scheduler/scheduler_ddim.ipynb.

# %% auto 0
__all__ = ['DDIMSchedulerOutput', 'DDIMScheduler']

# %% ../../src/scheduler/scheduler_ddim.ipynb 3
from ..imports import *
from .scheduler_ddpm import DDPMScheduler

# %% ../../src/scheduler/scheduler_ddim.ipynb 4
@dataclass
class DDIMSchedulerOutput:
    prev_sample: torch.FloatTensor
    pred_original_sample: Optional[torch.FloatTensor] = None
    
class DDIMScheduler(DDPMScheduler):
    """A `Scheduler` implementing [(DDIM)](https://arxiv.org/abs/2010.02502)."""
    def __init__(self, 
                 device: Union[str, torch.device],     
                 num_train_timesteps: int = 1000,
                 beta_start: float = 0.0001,
                 beta_end: float = 0.02,
                 beta_schedule: str = "linear",
                 input_perturbation = 0.1,
                 eta: float = 0
                ):    
        super().__init__(device, num_train_timesteps, beta_start, beta_end, beta_schedule, input_perturbation)
        self.eta = eta
        
        #for stable diff ddim
        set_alpha_to_one = True  # False 
        steps_offset     = 0     # 1
            
        self.steps_offset        = steps_offset
        self.final_alpha_cumprod = torch.tensor(1.0) if set_alpha_to_one else self.alphas_cumprod[0]
     
    @property
    def params_config(self):         
        params_config = super().params_config
        params_config["eta"] = self.eta
        return params_config
    
    #------------------------------------
    # Inference functions
    
    def set_timesteps(self, num_inference_steps: int):        
        super().set_timesteps(num_inference_steps)   
        self.timesteps += self.steps_offset
    
    clamp_style = None # one of: None, "static", "dynamic"
    
    def step(self,         
             model_output: torch.FloatTensor,
             timesteps: Union[int, torch.IntTensor],
             sample: torch.FloatTensor
            ) -> DDIMSchedulerOutput:
        """Denoising step"""
        
        prev_timesteps = timesteps - self.num_train_timesteps // self.num_inference_steps
        
        #get variance sched
        alphas_cumprod     = self.unsqueeze_vector_to_shape(self.alphas_cumprod[timesteps], sample.shape)
        alphas_cumprod_tm1 = self.unsqueeze_vector_to_shape(self.alphas_cumprod[prev_timesteps], sample.shape)
                       
        #fix negative timesteps to self.final_alpha_cumprod        
        non_zero_tm1 = (prev_timesteps>=0.0).float()      
        non_zero_tm1 = self.unsqueeze_vector_to_shape(non_zero_tm1, sample.shape)
        alphas_cumprod_tm1 = alphas_cumprod_tm1 * non_zero_tm1 + (1.0 - non_zero_tm1) *  self.final_alpha_cumprod  
                              
        #calc vars
        betas_cumprod     = 1.0 - alphas_cumprod
        betas_cumprod_tm1 = 1.0 - alphas_cumprod_tm1
            
        #estimate predicted sample
        x0 = (sample - betas_cumprod.sqrt() * model_output) / alphas_cumprod.sqrt()
        
        if   self.clamp_style == None: pass
        elif self.clamp_style == "static":  x0 = torch.clamp(x0, -1, 1)
        elif self.clamp_style == "dynamic": raise NotImplementedError("clamp_style == 'dynamic'")
        else: raise NotImplementedError("self.clamp_style has to be one of: None, 'static', 'dynamic'")       
        #if self.num_train_timesteps // self.num_inference_steps > 1: x0 = torch.clamp(x0, -1, 1)
    
        #variance
        variance = (betas_cumprod_tm1 / betas_cumprod) * (1.0 - alphas_cumprod / alphas_cumprod_tm1)
        std      = self.eta * variance.sqrt()
        
        #direction to xt
        dir_xt = ( betas_cumprod_tm1 - std.square() ).sqrt() * model_output
    
        #sample noise
        noise = torch.randn(model_output.shape, device=self.device)
    
        #estimate the prev sample      
        xtm1 = alphas_cumprod_tm1.sqrt() * x0 + dir_xt + std * noise
        
        # print(f"{timesteps=} {prev_timesteps=} ;;;   x0:  {x0.mean()}+-{x0.std()}      xtm1:  {xtm1.mean()}+-{xtm1.std()}")
        
        return DDIMSchedulerOutput(prev_sample=xtm1, pred_original_sample=x0)    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/scheduler/scheduler_ddpm.ipynb.

# %% auto 0
__all__ = ['DDPMSchedulerOutput', 'DDPMScheduler']

# %% ../../src/scheduler/scheduler_ddpm.ipynb 3
from ..imports import *
from .scheduler import Scheduler

# %% ../../src/scheduler/scheduler_ddpm.ipynb 4
@dataclass
class DDPMSchedulerOutput:
    prev_sample: torch.FloatTensor
    pred_original_sample: Optional[torch.FloatTensor] = None
    
class DDPMScheduler(Scheduler):
    """A `Scheduler` implementing [(DDPM)](https://arxiv.org/abs/2006.11239)"""
    
    non_blocking = False
    
    def __init__(self, 
                 device: Union[str, torch.device],     
                 num_train_timesteps: int = 1000,
                 beta_start: float = 0.0001,
                 beta_end: float = 0.02,
                 beta_schedule: str = "linear",
                 input_perturbation = 0.1
                ):    
        super().__init__()
        self.device = device        
        self.num_train_timesteps = torch.tensor(num_train_timesteps)
        self.num_inference_steps = torch.tensor(num_train_timesteps)
        
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule

        self.timesteps = torch.from_numpy(np.arange(0, num_train_timesteps)[::-1].copy().astype(np.int64)) #careful is defined reversed for easy denoising looping
            
        if beta_schedule == "linear":
            self.betas = torch.linspace(beta_start, beta_end, num_train_timesteps, dtype=torch.float32)
            
        elif beta_schedule == "linear_sqrt":
            self.betas = torch.linspace(beta_start ** 0.5, beta_end ** 0.5, num_train_timesteps, dtype=torch.float32) ** 2
                      
        elif beta_schedule == "cos_alpha":  #cosine-based-variance                      
            f = lambda t: np.cos((t/self.num_train_timesteps + 0.008)*np.pi/2.016)**2 
            _betas = []            
            for i in range(self.num_train_timesteps):                           
                _betas.append(min(1.0-(f(i+1.0)/f(i)),0.999))                
            self.betas = torch.tensor(_betas, dtype=torch.float32)                       
        else:
            raise NotImplementedError(f"{beta_schedule} does is not implemented for {self.__class__}")
              
        self.sigmas = torch.sqrt(self.betas)
                 
        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = torch.cumprod(self.alphas, dim=0)  
        
        ## Is this used anymore?
        self.sqrt_alphas                   = torch.sqrt(self.alphas)  
        self.sqrt_alphas_cumprod           = torch.sqrt(self.alphas_cumprod)
        self.sqrt_one_minus_alphas_cumprod = torch.sqrt(1 - self.alphas_cumprod)
        ##
        
        #----------
        # Input Perturbation Reduces Exposure Bias in Diffusion Models
        # https://arxiv.org/pdf/2301.11706.pdf
        
        self.input_perturbation = input_perturbation
        
        #----------
        
        self.to_device(self.device)
    
    @property
    def params_config(self):         
        params_config = {}           
        params_config["device"]              = str(self.device)
        params_config["num_train_timesteps"] = self.num_train_timesteps.item()
        params_config["beta_start"]          = self.beta_start
        params_config["beta_end"]            = self.beta_end
        params_config["beta_schedule"]       = self.beta_schedule
        params_config["input_perturbation"]  = self.input_perturbation            
        return params_config
    
    def to_device(self, device: Union[str, torch.device], non_blocking=False):
        #non_blocking = self.non_blocking
        
        self.device                        = device
        self.alphas_cumprod                = self.alphas_cumprod.to(device, non_blocking=non_blocking)
        self.sqrt_alphas_cumprod           = self.sqrt_alphas_cumprod.to(device, non_blocking=non_blocking)
        self.sqrt_one_minus_alphas_cumprod = self.sqrt_one_minus_alphas_cumprod.to(device, non_blocking=non_blocking)
        self.sigmas                        = self.sigmas.to(device, non_blocking=non_blocking)
        self.sqrt_alphas                   = self.sqrt_alphas.to(device, non_blocking=non_blocking)
        self.betas                         = self.betas.to(device, non_blocking=non_blocking)
        self.num_train_timesteps           = self.num_train_timesteps.to(device, non_blocking=non_blocking)
        self.num_inference_steps           = self.num_inference_steps.to(device, non_blocking=non_blocking)
        
    #------------------------------------
    # Inference functions
    
    def set_timesteps(self, num_inference_steps: int):        
        if num_inference_steps >= self.num_train_timesteps: raise ValueError("num_inference_steps >= self.num_train_timesteps")     
        self.num_inference_steps = torch.tensor(num_inference_steps)
        step_ratio = self.num_train_timesteps // self.num_inference_steps
        timesteps = (np.arange(0, num_inference_steps) * step_ratio.item()).round()[::-1].copy().astype(np.int64)
        self.timesteps = torch.from_numpy(timesteps)
               
    def step(self,
             model_output: torch.FloatTensor,
             timesteps: Union[int, torch.IntTensor],
             sample: torch.FloatTensor
            ) -> DDPMSchedulerOutput:
        """Denoising step"""
        
        sqrt_alphas_cumprod = self.unsqueeze_vector_to_shape(self.sqrt_alphas_cumprod[timesteps], sample.shape)
        sqrt_one_minus_alphas_cumprod = self.unsqueeze_vector_to_shape(self.sqrt_one_minus_alphas_cumprod[timesteps], sample.shape)
           
        sigmas      = self.unsqueeze_vector_to_shape(self.sigmas[timesteps], sample.shape)
        sqrt_alphas = self.unsqueeze_vector_to_shape(self.sqrt_alphas[timesteps], sample.shape)
        betas       = self.unsqueeze_vector_to_shape(self.betas[timesteps], sample.shape)
      
        non_zero_t = (timesteps!=0).float()
    
        #estimate the final img
        x0 = (sample - sqrt_one_minus_alphas_cumprod * model_output) / sqrt_alphas_cumprod   #DDPM eq.15
               
        #less noisy latent     
        noise = torch.randn(sample.shape, device=self.device) 
        noise = noise * non_zero_t.reshape(-1, 1, 1, 1)
        
        xt_coeff = betas / sqrt_one_minus_alphas_cumprod        
        xt = (sample - xt_coeff * model_output) / sqrt_alphas + sigmas * noise 
        
        return DDPMSchedulerOutput(prev_sample=xt, pred_original_sample=x0)

    def add_noise_LEdit(self, original_samples: torch.FloatTensor):
        # LEDITS: Real Image Editing with DDPM Inversion and Semantic Guidance; Note: SEGA (Semantic Guidance) is just multiple negative promts with a pixel based weight
        # https://arxiv.org/pdf/2307.00522.pdf

        noisy_latents = []
        noises        = []

        noisy_latent_t = original_samples
        
        for t in self.timesteps[::-1]: #start from no noise and diffuse in non analytic fashion
            noise_t        = torch.randn_like(noise)   
            alpha_t        = self.unsqueeze_vector_to_shape(self.alphas[t], original_samples.shape)    
            noisy_latent_t = torch.sqrt(alpha_t) * noisy_latent_t + torch.sqrt(1.0 - alpha_t) * noise_t       

            noises.append(noise_t)
            noisy_latents.append(noisy_latent_t)
        
        return noisy_latents[::-1], noises[::-1] # invert to self.timestep definition
    
    #------------------------------------
    # Training functions
        
    def add_noise(self,
                  original_samples: torch.FloatTensor,
                  noise: torch.FloatTensor,
                  timesteps: torch.IntTensor,
                 ) -> torch.FloatTensor:
            
        alphas_cumprod = self.unsqueeze_vector_to_shape(self.alphas_cumprod[timesteps], original_samples.shape)        
        noisy_latents = torch.sqrt(alphas_cumprod) * original_samples + torch.sqrt(1.0 - alphas_cumprod) * noise       
        
        if self.input_perturbation is not None:
            noisy_latents = noisy_latents + torch.sqrt(1.0 - alphas_cumprod) * torch.randn_like(noise) * self.input_perturbation
        
        return noisy_latents  



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/platform/qcircuit_evaluation.ipynb.

# %% auto 0
__all__ = ['sort_into_bins', 'extract_gate_number', 'get_gate_stat_from_tensors', 'get_gate_stat_from_circuits']

# %% ../../src/platform/qcircuit_evaluation.ipynb 2
from ..imports import *
from .qcircuit_dataset_construction import *
from .simulation.qcircuit_sim import schmidt_rank_vector, optimize_circuit

import qiskit.quantum_info as qi
from qiskit import QuantumCircuit

# %% ../../src/platform/qcircuit_evaluation.ipynb 4
def sort_into_bins(x, y, y_uniques):
    
    x_binned = []
    y_binned = []
    
    for y_unique in y_uniques:
    
        comp    = torch.all(y==y_unique, dim=-1)
        indices = comp.nonzero().squeeze()
        
        x_binned.append(x[indices])
        y_binned.append(y[indices])
    
    y_bins = [y[0] for y in y_binned]
    
    return x_binned, y_binned, y_bins

# %% ../../src/platform/qcircuit_evaluation.ipynb 5
def extract_gate_number(qc: QuantumCircuit, gate_pool, max_gates):    
    gate_classes = {"empty":0} | {x().name:i+1 for i,x in enumerate(gate_pool)}
       
    gate_cnt = np.zeros(len(gate_classes), dtype=int)   
    
    if hasattr(qc, "data"):    
        for t, gate in enumerate(qc.data):   
            gate_id = gate_classes[gate.operation.name]       
            gate_cnt[gate_id] += 1
                
    gate_cnt[0] = max_gates - sum(gate_cnt[1:])
        
    return gate_cnt, gate_classes

# %% ../../src/platform/qcircuit_evaluation.ipynb 6
def get_gate_stat_from_tensors(tensors, gate_pool):
    for i,tensor in tqdm(enumerate(tensors), total=tensors.shape[0]):       
        qc = decode_circuit(tensor, gate_pool)
        
        t_gate_cnts, gate_dict = extract_gate_number(qc, gate_pool, max_gates=tensor.shape[1])
  
        if i > 0: gate_cnts = np.vstack([gate_cnts, t_gate_cnts])
        else:     gate_cnts = t_gate_cnts

    return gate_cnts, gate_dict

# %% ../../src/platform/qcircuit_evaluation.ipynb 7
def get_gate_stat_from_circuits(qcs: list, gate_pool, max_gates):
    for i,qc in tqdm(enumerate(qcs), total=len(qcs)):
        
        t_gate_cnts, gate_dict = extract_gate_number(qc, gate_pool, max_gates)
  
        if i > 0: gate_cnts = np.vstack([gate_cnts, t_gate_cnts])
        else:     gate_cnts = t_gate_cnts

    return gate_cnts, gate_dict



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/pipeline/diffusion_pipeline_special.ipynb.

# %% auto 0
__all__ = ['DiffusionPipeline_attnPadded', 'DiffusionPipeline_Compilation']

# %% ../../src/pipeline/diffusion_pipeline_special.ipynb 2
from ..imports import *
from .diffusion_pipeline import DiffusionPipeline

# %% ../../src/pipeline/diffusion_pipeline_special.ipynb 3
class DiffusionPipeline_attnPadded(DiffusionPipeline): 
    """A special `DiffusionPipeline` with attention masking."""
    def train_step(self, data, **kwargs): 
        latents, y, key_padding_mask_list = data                
        b, s, t = latents.shape          
        
        #start async memcpy
        loss_mask = (key_padding_mask_list[0].to(self.device, non_blocking=self.non_blocking)>-1.0).float().unsqueeze(1)
      
        shaped_mask = []
        for key_padding_mask in key_padding_mask_list:
            key_padding_mask = key_padding_mask.to(self.device, non_blocking=self.non_blocking)    
            key_padding_mask = key_padding_mask.reshape((b, -1)) #from [b, s, t] to [b, -1] aka [N, L]           
            shaped_mask.append(key_padding_mask)         
        
        latents = latents.to(self.device, non_blocking=self.non_blocking)  
        latents = self.model.embedd_clrs(latents)                   #this is only new tensor
        self.scheduler.to_device(self.device, non_blocking=self.non_blocking) 
            
        #do the cond embedding with CLIP                     
        y = y.to(self.device, non_blocking=self.non_blocking)  
        
        if self.enable_guidance_train: 
            rnd = torch.rand((b,), device=self.device)            
            rnd = (rnd > self.guidance_train_p).type(torch.int64)          # todo: change to bernoulli dist fn
            rnd = self.scheduler.unsqueeze_vector_to_shape(rnd, y.shape)   # e.g. [b, 1, 1]            
            y   = y * rnd + (1-rnd) * self.empty_token_fn(y)
               
        y_emb = self.text_encoder(y, pool=False)
                          
        #sample timesteps
        timesteps = torch.randint(low=0, high=self.scheduler.num_train_timesteps, size=(b,), device=self.device, dtype=torch.int64)

        #forward noising    
        noise = torch.randn(latents.shape, device=self.device)     
        noisy_latents = self.scheduler.add_noise(latents, noise, timesteps)

        #predict eps
        eps = self.model(noisy_latents, timesteps, y_emb, key_padding_mask=shaped_mask)
            
        #comp mse
        loss = self.loss_fn(eps*loss_mask, noise*loss_mask)
        # loss = self.loss_fn(eps, noise)
        
        #log the loss
        return loss

# %% ../../src/pipeline/diffusion_pipeline_special.ipynb 4
class DiffusionPipeline_Compilation(DiffusionPipeline):   
    """A special `DiffusionPipeline` that accounts for unitary conditions, i.e. compilation."""
    
    #------------------------------------
    
    @torch.no_grad()
    def __call__(self, latents, c, U, g, no_bar=False):
        
        latents = latents.to(self.device)
        c       = c.to(self.device)
        U       = U.to(self.device)
        
        return self.denoising(latents, c=c, U=U, enable_guidance=True, g=g, no_bar=no_bar)

    #------------------------------------
    
    def get_guidance_U(self, U, enable_guidance):
        if not exists(U): return U      
        U = U.to(self.device)                
        if enable_guidance:             
            u = torch.zeros_like(U, device=self.device)       
            U = torch.cat([u, U])            
        U = U.type(torch.float32)
        return U
    
    @torch.no_grad()
    def denoising(self, latents, c, U, enable_guidance, g, no_bar=False, return_predicted_x0=False):       
        U = self.get_guidance_U(U, enable_guidance)  
        # self.unitary_encoder.eval()   
        return super().denoising(latents, c, enable_guidance, g, no_bar=no_bar, return_predicted_x0=return_predicted_x0, U=U)

    def denoising_step(self, latents: torch.Tensor, ts: Union[int, torch.IntTensor], c_emb: torch.Tensor=None, enable_guidance=False, g=7.5, U: torch.Tensor=None):    
        if enable_guidance:
            x = torch.cat([latents] * 2)     #uses batch layer combine here
            
            if ts.numel() > 1: chunk_ts = torch.cat([ts] * 2)
            else:              chunk_ts = ts
                
            eps_u, eps_c = self.model(x, chunk_ts, c_emb, U=U).chunk(2) 
            
            eps = self.CFG(eps_u, eps_c, g)
                    
        else:
            eps = self.model(latents, ts, c_em, U=U)  
                 
        x = self.scheduler.step(eps, ts, latents)      
        return x.prev_sample, x.pred_original_sample
    
    #------------------------------------

    def train_step(self, data, **kwargs): 
        latents, y, U = data                
        b, s, t = latents.shape          
        
        #start async memcpy
        latents = latents.to(self.device, non_blocking=self.non_blocking)  
        latents = self.model.embedd_clrs(latents)                   #this is only new tensor
            
        #do the cond embedding with CLIP                     
        y = y.to(self.device, non_blocking=self.non_blocking)  
        U = U.to(self.device, non_blocking=self.non_blocking)  
        
        if self.enable_guidance_train: 
            rnd = torch.rand((b,), device=self.device)            
            rnd = (rnd > self.guidance_train_p).type(torch.int64)          # todo: change to bernoulli dist fn
            
            rnd_y = self.scheduler.unsqueeze_vector_to_shape(rnd, y.shape)   # e.g. [b, 1, 1]            
            y   = y * rnd_y + (1-rnd_y) * self.empty_token_fn(y)
            
            U   = U * self.scheduler.unsqueeze_vector_to_shape(rnd, U.shape)
            
        y_emb = self.text_encoder(y, pool=False)
              
        #sample timesteps
        timesteps = torch.randint(low=0, high=self.scheduler.num_train_timesteps, size=(b,), device=self.device, dtype=torch.int64)

        #forward noising    
        noise = torch.randn(latents.shape, device=self.device)     
        noisy_latents = self.scheduler.add_noise(latents, noise, timesteps)

        #predict eps
        eps = self.model(noisy_latents, timesteps, y_emb, U=U)
            
        #comp mse
        loss = self.loss_fn(eps, noise)
        
        #log the loss
        return loss



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../src/platform/simulation/qcircuit_sim.ipynb.

# %% auto 0
__all__ = ['get_number_of_gate_params', 'gate_pool_to_gate_classes', 'instruction_name_to_qiskit_gate', 'schmidt_rank_vector',
           'rnd_circuit', 'optimize_circuit', 'plot_svr_stat']

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 2
from ...imports import *
from ...config_loader import *

import qiskit.quantum_info as qi
from qiskit import QuantumCircuit, transpile
from qiskit.circuit.gate import Gate
import qiskit.circuit.library as ql

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 4
def get_number_of_gate_params(gate_cls):
    return gate_cls.__init__.__code__.co_argcount - len(gate_cls.__init__.__defaults__) - 1 # python: gives you the number of any arguments BEFORE *args, minus ones that have a default, -1 for self parameter of classes

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 5
def gate_pool_to_gate_classes(gate_pool: list[Gate]): 
    """Creates a vocabulary from a gate pool."""
    classes = {}
    
    for i,cls in enumerate(gate_pool):
        num_of_paramters = get_number_of_gate_params(cls)
        name = cls(*[0]*num_of_paramters).name
        classes[name] = (i+1)
        
    return classes

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 6
def instruction_name_to_qiskit_gate(name: str) -> Gate:
    match name:
        case "swap": name = "Swap"
        case "cp":   name = "CPhase"
        case _:      name = name.upper()
        
    return get_obj_from_str(f"qiskit.circuit.library.standard_gates.{name}Gate")

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 7
def schmidt_rank_vector(densityMatrix: qi.DensityMatrix):   
    """Return the SRV of a `qi.DensityMatrix`."""
    systems_cnt = len(densityMatrix.dims())   
    total_trace = set(range(systems_cnt))    
    rank_vector = []
    
    for i in range(systems_cnt): 
        trace = list(total_trace - {i})
        red_densityMatrix = qi.partial_trace(densityMatrix, trace)        
        # r = np.count_nonzero(np.linalg.eigvals(red_densityMatrix) > 1e-14) # was slower during testing   
        r = np.linalg.matrix_rank(red_densityMatrix, hermitian=True).item()       
        rank_vector.append(r)
    
    return rank_vector

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 8
def rnd_circuit(num_of_qubits, num_of_gates, gate_pool: list[Gate], rng):
    """Create a random circuit."""
    qc = QuantumCircuit(num_of_qubits)    
    gate_indices = rng.choice(len(gate_pool), num_of_gates)
    
    for gate_index in gate_indices:
        gate_qiskit_class = gate_pool[gate_index]
        
        num_of_paramters = get_number_of_gate_params(gate_qiskit_class)
        params           = rng.uniform(low=0, high=2*np.pi, size=num_of_paramters) if num_of_paramters > 0 else [] # random between 0 and 2pi
        
        gate = gate_qiskit_class(*params)      
        act_qubits = rng.choice(num_of_qubits, gate.num_qubits, replace=False) # order: (*act_qubits)=(*control_qubits, *target_qubits)   
        qc.append(gate, [*act_qubits], [])
    
    return qc

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 9
def optimize_circuit(qc: QuantumCircuit, gate_pool: list[Gate], optimization_level=2):
    """Use qiskit.compiler.transpile to optimize a circuit."""
    basis_gates = gate_pool_to_gate_classes(gate_pool).keys()
    
    while optimization_level > 0:
        try:
            qc_opt = transpile(qc, optimization_level=optimization_level, basis_gates=basis_gates) #target=target
            return qc_opt
        except Exception as er: pass
  
        optimization_level -= 1

    return qc

# %% ../../../src/platform/simulation/qcircuit_sim.ipynb 11
def plot_svr_stat(num_of_qubits, min_gates, max_gates, gs, samples, sort=False, opt=True, rng=np.random.default_rng()):    
    svr_list = list()
    for i in range(samples):
        qc = rnd_circuit(num_of_qubits, rng.integers(min_gates, max_gates+1), gs, rng) 
        if opt: qc = optimize_circuit(qc, gs)
        svr = schmidt_rank_vector(qi.DensityMatrix(qc))
        if sort: svr = sorted(svr)
        svr_list.append(svr)           
    df = pd.DataFrame(data={"svr":svr_list})   
    cnts = df['svr'].value_counts(normalize=True)
    for n,v in zip(cnts.index, cnts.values): print(f"{n}: {v*100:.1f}%")    
    df['svr'].value_counts().plot(kind='bar')   



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/platform/qcircuit_metrics.ipynb.

# %% auto 0
__all__ = ['Unitary_FrobeniusNorm']

# %% ../../src/platform/qcircuit_metrics.ipynb 3
from ..imports import *

# %% ../../src/platform/qcircuit_metrics.ipynb 5
class Unitary_FrobeniusNorm:
    #defined in https://arxiv.org/pdf/2106.05649.pdf
    
    @staticmethod
    def distance(approx_U: torch.tensor, target_U: torch.tensor):
        d = 0.5 * torch.linalg.matrix_norm((approx_U-target_U), ord="fro")**2
        return d
        
    @staticmethod
    def name():
        return "Frobenius-Norm"



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/genQC',
                'doc_host': 'https://FlorianFuerrutter.github.io',
                'git_url': 'https://github.com/FlorianFuerrutter/genQC',
                'lib_path': 'genQC'},
  'syms': { 'genQC.config_loader': { 'genQC.config_loader.class_to_str': ('config_loader.html#class_to_str', 'genQC/config_loader.py'),
                                     'genQC.config_loader.config_to_dict': ('config_loader.html#config_to_dict', 'genQC/config_loader.py'),
                                     'genQC.config_loader.get_obj_from_str': ( 'config_loader.html#get_obj_from_str',
                                                                               'genQC/config_loader.py'),
                                     'genQC.config_loader.instantiate_from_config': ( 'config_loader.html#instantiate_from_config',
                                                                                      'genQC/config_loader.py'),
                                     'genQC.config_loader.load_config': ('config_loader.html#load_config', 'genQC/config_loader.py'),
                                     'genQC.config_loader.load_model_from_config': ( 'config_loader.html#load_model_from_config',
                                                                                     'genQC/config_loader.py'),
                                     'genQC.config_loader.save_dataclass_yaml': ( 'config_loader.html#save_dataclass_yaml',
                                                                                  'genQC/config_loader.py'),
                                     'genQC.config_loader.save_dict_yaml': ('config_loader.html#save_dict_yaml', 'genQC/config_loader.py')},
            'genQC.dataset.cached_qc_dataset': { 'genQC.dataset.cached_qc_dataset.Cached_OpenClip_Dataset': ( 'dataset/cached_qc_dataset.html#cached_openclip_dataset',
                                                                                                              'genQC/dataset/cached_qc_dataset.py'),
                                                 'genQC.dataset.cached_qc_dataset.Cached_OpenClip_Dataset.caching': ( 'dataset/cached_qc_dataset.html#cached_openclip_dataset.caching',
                                                                                                                      'genQC/dataset/cached_qc_dataset.py'),
                                                 'genQC.dataset.cached_qc_dataset.Cached_OpenClip_Dataset.from_config_file': ( 'dataset/cached_qc_dataset.html#cached_openclip_dataset.from_config_file',
                                                                                                                               'genQC/dataset/cached_qc_dataset.py'),
                                                 'genQC.dataset.cached_qc_dataset.Cached_OpenClip_Dataset.get_dataloaders': ( 'dataset/cached_qc_dataset.html#cached_openclip_dataset.get_dataloaders',
                                                                                                                              'genQC/dataset/cached_qc_dataset.py'),
                                                 'genQC.dataset.cached_qc_dataset.Cached_OpenClip_Dataset.x_y_preprocess': ( 'dataset/cached_qc_dataset.html#cached_openclip_dataset.x_y_preprocess',
                                                                                                                             'genQC/dataset/cached_qc_dataset.py')},
            'genQC.dataset.config_dataset': { 'genQC.dataset.config_dataset.Config_Dataset': ( 'dataset/config_dataset.html#config_dataset',
                                                                                               'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.__init__': ( 'dataset/config_dataset.html#config_dataset.__init__',
                                                                                                        'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.from_config': ( 'dataset/config_dataset.html#config_dataset.from_config',
                                                                                                           'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.from_config_file': ( 'dataset/config_dataset.html#config_dataset.from_config_file',
                                                                                                                'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.from_huggingface': ( 'dataset/config_dataset.html#config_dataset.from_huggingface',
                                                                                                                'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.get_config': ( 'dataset/config_dataset.html#config_dataset.get_config',
                                                                                                          'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.load_x_y': ( 'dataset/config_dataset.html#config_dataset.load_x_y',
                                                                                                        'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.params_config': ( 'dataset/config_dataset.html#config_dataset.params_config',
                                                                                                             'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.save_dataset': ( 'dataset/config_dataset.html#config_dataset.save_dataset',
                                                                                                            'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.store_x_y': ( 'dataset/config_dataset.html#config_dataset.store_x_y',
                                                                                                         'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset.to': ( 'dataset/config_dataset.html#config_dataset.to',
                                                                                                  'genQC/dataset/config_dataset.py'),
                                              'genQC.dataset.config_dataset.Config_Dataset_config': ( 'dataset/config_dataset.html#config_dataset_config',
                                                                                                      'genQC/dataset/config_dataset.py')},
            'genQC.dataset.dataset_helper': { 'genQC.dataset.dataset_helper.balance_tensor_dataset': ( 'dataset/dataset_helper.html#balance_tensor_dataset',
                                                                                                       'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.check_duplicate_in_dataset': ( 'dataset/dataset_helper.html#check_duplicate_in_dataset',
                                                                                                           'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.check_duplicates_in_dataset': ( 'dataset/dataset_helper.html#check_duplicates_in_dataset',
                                                                                                            'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.check_duplicates_in_dataset_python': ( 'dataset/dataset_helper.html#check_duplicates_in_dataset_python',
                                                                                                                   'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.get_unique_elements_indices': ( 'dataset/dataset_helper.html#get_unique_elements_indices',
                                                                                                            'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.map_old_tensor_to_new': ( 'dataset/dataset_helper.html#map_old_tensor_to_new',
                                                                                                      'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.shuffle_tensor_dataset': ( 'dataset/dataset_helper.html#shuffle_tensor_dataset',
                                                                                                       'genQC/dataset/dataset_helper.py'),
                                              'genQC.dataset.dataset_helper.uniquify_tensor_dataset': ( 'dataset/dataset_helper.html#uniquify_tensor_dataset',
                                                                                                        'genQC/dataset/dataset_helper.py')},
            'genQC.dataset.mixed_cached_qc_dataset': { 'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset',
                                                                                                                                'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_bucket_collate_fn',
                                                                                                                                                              'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn_compilation': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_bucket_collate_fn_compilation',
                                                                                                                                                                          'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn_compilation_params': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_bucket_collate_fn_compilation_params',
                                                                                                                                                                                 'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_collate_fn',
                                                                                                                                                       'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn_compilation': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_collate_fn_compilation',
                                                                                                                                                                   'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn_compilation_params': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.cut_padding_collate_fn_compilation_params',
                                                                                                                                                                          'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.flexPadAttn_TimeOnly_padding_collate_fn': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.flexpadattn_timeonly_padding_collate_fn',
                                                                                                                                                                        'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.flexPadAttn_padding_collate_fn': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.flexpadattn_padding_collate_fn',
                                                                                                                                                               'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.from_config_file': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.from_config_file',
                                                                                                                                                 'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.from_datasets': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.from_datasets',
                                                                                                                                              'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.get_dataloaders': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.get_dataloaders',
                                                                                                                                                'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset.params_config': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset.params_config',
                                                                                                                                              'genQC/dataset/mixed_cached_qc_dataset.py'),
                                                       'genQC.dataset.mixed_cached_qc_dataset.Mixed_Cached_OpenClip_Dataset_config': ( 'dataset/mixed_cached_qc_dataset.html#mixed_cached_openclip_dataset_config',
                                                                                                                                       'genQC/dataset/mixed_cached_qc_dataset.py')},
            'genQC.dataset.qc_dataset': { 'genQC.dataset.qc_dataset.Qc_Config_Dataset': ( 'dataset/qc_dataset.html#qc_config_dataset',
                                                                                          'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.__init__': ( 'dataset/qc_dataset.html#qc_config_dataset.__init__',
                                                                                                   'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.get_dataloaders': ( 'dataset/qc_dataset.html#qc_config_dataset.get_dataloaders',
                                                                                                          'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.params_config': ( 'dataset/qc_dataset.html#qc_config_dataset.params_config',
                                                                                                        'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.plot_distribution': ( 'dataset/qc_dataset.html#qc_config_dataset.plot_distribution',
                                                                                                            'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.plot_example': ( 'dataset/qc_dataset.html#qc_config_dataset.plot_example',
                                                                                                       'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.valid_split': ( 'dataset/qc_dataset.html#qc_config_dataset.valid_split',
                                                                                                      'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset.x_y_preprocess': ( 'dataset/qc_dataset.html#qc_config_dataset.x_y_preprocess',
                                                                                                         'genQC/dataset/qc_dataset.py'),
                                          'genQC.dataset.qc_dataset.Qc_Config_Dataset_config': ( 'dataset/qc_dataset.html#qc_config_dataset_config',
                                                                                                 'genQC/dataset/qc_dataset.py')},
            'genQC.imports': {},
            'genQC.inference.export_cudaq': { 'genQC.inference.export_cudaq.CircuitInstruction': ( 'inference/export_cudaq.html#circuitinstruction',
                                                                                                   'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions': ( 'inference/export_cudaq.html#circuitinstructions',
                                                                                                    'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.__init__': ( 'inference/export_cudaq.html#circuitinstructions.__init__',
                                                                                                             'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.__repr__': ( 'inference/export_cudaq.html#circuitinstructions.__repr__',
                                                                                                             'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.add_instruction': ( 'inference/export_cudaq.html#circuitinstructions.add_instruction',
                                                                                                                    'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.data': ( 'inference/export_cudaq.html#circuitinstructions.data',
                                                                                                         'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.length': ( 'inference/export_cudaq.html#circuitinstructions.length',
                                                                                                           'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.max_gates': ( 'inference/export_cudaq.html#circuitinstructions.max_gates',
                                                                                                              'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.num_qubits': ( 'inference/export_cudaq.html#circuitinstructions.num_qubits',
                                                                                                               'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitInstructions.print': ( 'inference/export_cudaq.html#circuitinstructions.print',
                                                                                                          'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend': ( 'inference/export_cudaq.html#circuitscudaqbackend',
                                                                                                     'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend._construct_kernel': ( 'inference/export_cudaq.html#circuitscudaqbackend._construct_kernel',
                                                                                                                       'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend.check_error_circuit': ( 'inference/export_cudaq.html#circuitscudaqbackend.check_error_circuit',
                                                                                                                         'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend.draw': ( 'inference/export_cudaq.html#circuitscudaqbackend.draw',
                                                                                                          'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend.export_cudaq': ( 'inference/export_cudaq.html#circuitscudaqbackend.export_cudaq',
                                                                                                                  'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.CircuitsCudaqBackend.get_unitary': ( 'inference/export_cudaq.html#circuitscudaqbackend.get_unitary',
                                                                                                                 'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.genqc_to_cudaq': ( 'inference/export_cudaq.html#genqc_to_cudaq',
                                                                                               'genQC/inference/export_cudaq.py'),
                                              'genQC.inference.export_cudaq.tensor_to_instructions': ( 'inference/export_cudaq.html#tensor_to_instructions',
                                                                                                       'genQC/inference/export_cudaq.py')},
            'genQC.inference.infer_compilation': { 'genQC.inference.infer_compilation.check_correct_gates': ( 'inference/infer_compilation.html#check_correct_gates',
                                                                                                              'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.check_correct_unitary_distance': ( 'inference/infer_compilation.html#check_correct_unitary_distance',
                                                                                                                         'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.check_correct_unitary_exact': ( 'inference/infer_compilation.html#check_correct_unitary_exact',
                                                                                                                      'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.generate_comp_tensors': ( 'inference/infer_compilation.html#generate_comp_tensors',
                                                                                                                'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.get_gate_and_U_acc': ( 'inference/infer_compilation.html#get_gate_and_u_acc',
                                                                                                             'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.get_new_unitary_indices': ( 'inference/infer_compilation.html#get_new_unitary_indices',
                                                                                                                  'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.get_new_unitary_indices_batch': ( 'inference/infer_compilation.html#get_new_unitary_indices_batch',
                                                                                                                        'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.plot_hist_overview': ( 'inference/infer_compilation.html#plot_hist_overview',
                                                                                                             'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.split_U_to_tensor': ( 'inference/infer_compilation.html#split_u_to_tensor',
                                                                                                            'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.test_comp_acc': ( 'inference/infer_compilation.html#test_comp_acc',
                                                                                                        'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.test_comp_acc_on_rnd_samples': ( 'inference/infer_compilation.html#test_comp_acc_on_rnd_samples',
                                                                                                                       'genQC/inference/infer_compilation.py'),
                                                   'genQC.inference.infer_compilation.test_comp_acc_on_testset': ( 'inference/infer_compilation.html#test_comp_acc_on_testset',
                                                                                                                   'genQC/inference/infer_compilation.py')},
            'genQC.inference.infer_gate_hist': { 'genQC.inference.infer_gate_hist.get_circuit_gate_length': ( 'inference/infer_gate_hist.html#get_circuit_gate_length',
                                                                                                              'genQC/inference/infer_gate_hist.py'),
                                                 'genQC.inference.infer_gate_hist.get_tensor_gate_length': ( 'inference/infer_gate_hist.html#get_tensor_gate_length',
                                                                                                             'genQC/inference/infer_gate_hist.py')},
            'genQC.inference.infer_misc': { 'genQC.inference.infer_misc.convert_tensors_to_circuits': ( 'inference/infer_misc.html#convert_tensors_to_circuits',
                                                                                                        'genQC/inference/infer_misc.py'),
                                            'genQC.inference.infer_misc.get_rnd_gatepool_subset': ( 'inference/infer_misc.html#get_rnd_gatepool_subset',
                                                                                                    'genQC/inference/infer_misc.py')},
            'genQC.inference.infer_srv': { 'genQC.inference.infer_srv.convert_tensors_to_srvs': ( 'inference/infer_srv.html#convert_tensors_to_srvs',
                                                                                                  'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.generate_srv_tensors': ( 'inference/infer_srv.html#generate_srv_tensors',
                                                                                               'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.get_all_srvs': ( 'inference/infer_srv.html#get_all_srvs',
                                                                                       'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.get_srv_accuracy': ( 'inference/infer_srv.html#get_srv_accuracy',
                                                                                           'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.plot_guidance_dep': ( 'inference/infer_srv.html#plot_guidance_dep',
                                                                                            'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.plot_srv_acc_vs_length': ( 'inference/infer_srv.html#plot_srv_acc_vs_length',
                                                                                                 'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.plot_srv_acc_vs_maxLength': ( 'inference/infer_srv.html#plot_srv_acc_vs_maxlength',
                                                                                                    'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.plot_srv_clr_distribution_bin_accuracy': ( 'inference/infer_srv.html#plot_srv_clr_distribution_bin_accuracy',
                                                                                                                 'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.plot_srv_clr_distribution_hist': ( 'inference/infer_srv.html#plot_srv_clr_distribution_hist',
                                                                                                         'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_guidance_dep': ( 'inference/infer_srv.html#test_guidance_dep',
                                                                                            'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_srv_acc_vs_length': ( 'inference/infer_srv.html#test_srv_acc_vs_length',
                                                                                                 'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_srv_acc_vs_maxLength': ( 'inference/infer_srv.html#test_srv_acc_vs_maxlength',
                                                                                                    'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_srv_clr_distribution': ( 'inference/infer_srv.html#test_srv_clr_distribution',
                                                                                                    'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_srv_clr_distribution_bin_samples': ( 'inference/infer_srv.html#test_srv_clr_distribution_bin_samples',
                                                                                                                'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.test_srv_length_distribution': ( 'inference/infer_srv.html#test_srv_length_distribution',
                                                                                                       'genQC/inference/infer_srv.py'),
                                           'genQC.inference.infer_srv.true_sample_bin_dist': ( 'inference/infer_srv.html#true_sample_bin_dist',
                                                                                               'genQC/inference/infer_srv.py')},
            'genQC.metrics': { 'genQC.metrics.Accuracy': ('metrics.html#accuracy', 'genQC/metrics.py'),
                               'genQC.metrics.Accuracy._eval': ('metrics.html#accuracy._eval', 'genQC/metrics.py'),
                               'genQC.metrics.Mean': ('metrics.html#mean', 'genQC/metrics.py'),
                               'genQC.metrics.Mean.__init__': ('metrics.html#mean.__init__', 'genQC/metrics.py'),
                               'genQC.metrics.Mean._eval': ('metrics.html#mean._eval', 'genQC/metrics.py'),
                               'genQC.metrics.Mean.reset_state': ('metrics.html#mean.reset_state', 'genQC/metrics.py'),
                               'genQC.metrics.Mean.result': ('metrics.html#mean.result', 'genQC/metrics.py'),
                               'genQC.metrics.Mean.update_state': ('metrics.html#mean.update_state', 'genQC/metrics.py'),
                               'genQC.metrics.Metric': ('metrics.html#metric', 'genQC/metrics.py'),
                               'genQC.metrics.Metric.__init__': ('metrics.html#metric.__init__', 'genQC/metrics.py'),
                               'genQC.metrics.Metric.__repr__': ('metrics.html#metric.__repr__', 'genQC/metrics.py'),
                               'genQC.metrics.Metric._eval': ('metrics.html#metric._eval', 'genQC/metrics.py'),
                               'genQC.metrics.Metric.reset_state': ('metrics.html#metric.reset_state', 'genQC/metrics.py'),
                               'genQC.metrics.Metric.result': ('metrics.html#metric.result', 'genQC/metrics.py'),
                               'genQC.metrics.Metric.update_state': ('metrics.html#metric.update_state', 'genQC/metrics.py')},
            'genQC.models.config_model': { 'genQC.models.config_model.Config_Model': ( 'models/config_model.html#config_model',
                                                                                       'genQC/models/config_model.py'),
                                           'genQC.models.config_model.Config_Model.__init__': ( 'models/config_model.html#config_model.__init__',
                                                                                                'genQC/models/config_model.py'),
                                           'genQC.models.config_model.Config_Model.from_config': ( 'models/config_model.html#config_model.from_config',
                                                                                                   'genQC/models/config_model.py'),
                                           'genQC.models.config_model.Config_Model.from_config_file': ( 'models/config_model.html#config_model.from_config_file',
                                                                                                        'genQC/models/config_model.py'),
                                           'genQC.models.config_model.Config_Model.get_config': ( 'models/config_model.html#config_model.get_config',
                                                                                                  'genQC/models/config_model.py'),
                                           'genQC.models.config_model.Config_Model.store_model': ( 'models/config_model.html#config_model.store_model',
                                                                                                   'genQC/models/config_model.py')},
            'genQC.models.frozen_open_clip': { 'genQC.models.frozen_open_clip.CachedFrozenOpenCLIPEmbedder': ( 'models/frozen_open_clip.html#cachedfrozenopenclipembedder',
                                                                                                               'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.CachedFrozenOpenCLIPEmbedder.forward': ( 'models/frozen_open_clip.html#cachedfrozenopenclipembedder.forward',
                                                                                                                       'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.CachedFrozenOpenCLIPEmbedder.generate_cache': ( 'models/frozen_open_clip.html#cachedfrozenopenclipembedder.generate_cache',
                                                                                                                              'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.CachedFrozenOpenCLIPEmbedder.look_up_cos_sim_cached_index': ( 'models/frozen_open_clip.html#cachedfrozenopenclipembedder.look_up_cos_sim_cached_index',
                                                                                                                                            'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder': ( 'models/frozen_open_clip.html#frozenopenclipembedder',
                                                                                                         'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.__init__': ( 'models/frozen_open_clip.html#frozenopenclipembedder.__init__',
                                                                                                                  'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.encode_with_transformer': ( 'models/frozen_open_clip.html#frozenopenclipembedder.encode_with_transformer',
                                                                                                                                 'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.forward': ( 'models/frozen_open_clip.html#frozenopenclipembedder.forward',
                                                                                                                 'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.freeze': ( 'models/frozen_open_clip.html#frozenopenclipembedder.freeze',
                                                                                                                'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.from_config': ( 'models/frozen_open_clip.html#frozenopenclipembedder.from_config',
                                                                                                                     'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.get_config': ( 'models/frozen_open_clip.html#frozenopenclipembedder.get_config',
                                                                                                                    'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.store_model': ( 'models/frozen_open_clip.html#frozenopenclipembedder.store_model',
                                                                                                                     'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.text_transformer_forward': ( 'models/frozen_open_clip.html#frozenopenclipembedder.text_transformer_forward',
                                                                                                                                  'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.to': ( 'models/frozen_open_clip.html#frozenopenclipembedder.to',
                                                                                                            'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder.tokenize_and_push_to_device': ( 'models/frozen_open_clip.html#frozenopenclipembedder.tokenize_and_push_to_device',
                                                                                                                                     'genQC/models/frozen_open_clip.py'),
                                               'genQC.models.frozen_open_clip.FrozenOpenCLIPEmbedder_config': ( 'models/frozen_open_clip.html#frozenopenclipembedder_config',
                                                                                                                'genQC/models/frozen_open_clip.py')},
            'genQC.models.layers': { 'genQC.models.layers.DownBlock2D': ('models/layers.html#downblock2d', 'genQC/models/layers.py'),
                                     'genQC.models.layers.DownBlock2D.__init__': ( 'models/layers.html#downblock2d.__init__',
                                                                                   'genQC/models/layers.py'),
                                     'genQC.models.layers.DownBlock2D.forward': ( 'models/layers.html#downblock2d.forward',
                                                                                  'genQC/models/layers.py'),
                                     'genQC.models.layers.FeedForward': ('models/layers.html#feedforward', 'genQC/models/layers.py'),
                                     'genQC.models.layers.FeedForward.__init__': ( 'models/layers.html#feedforward.__init__',
                                                                                   'genQC/models/layers.py'),
                                     'genQC.models.layers.FeedForward.forward': ( 'models/layers.html#feedforward.forward',
                                                                                  'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding': ( 'models/layers.html#positionalencoding',
                                                                                 'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding.__init__': ( 'models/layers.html#positionalencoding.__init__',
                                                                                          'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding.forward': ( 'models/layers.html#positionalencoding.forward',
                                                                                         'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2D': ( 'models/layers.html#positionalencoding2d',
                                                                                   'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2D.__init__': ( 'models/layers.html#positionalencoding2d.__init__',
                                                                                            'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2D.forward': ( 'models/layers.html#positionalencoding2d.forward',
                                                                                           'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2DSpaceOnly': ( 'models/layers.html#positionalencoding2dspaceonly',
                                                                                            'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2DSpaceOnly.__init__': ( 'models/layers.html#positionalencoding2dspaceonly.__init__',
                                                                                                     'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncoding2DSpaceOnly.forward': ( 'models/layers.html#positionalencoding2dspaceonly.forward',
                                                                                                    'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncodingTransposed': ( 'models/layers.html#positionalencodingtransposed',
                                                                                           'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncodingTransposed.__init__': ( 'models/layers.html#positionalencodingtransposed.__init__',
                                                                                                    'genQC/models/layers.py'),
                                     'genQC.models.layers.PositionalEncodingTransposed.forward': ( 'models/layers.html#positionalencodingtransposed.forward',
                                                                                                   'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D': ('models/layers.html#resblock2d', 'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D.__init__': ( 'models/layers.html#resblock2d.__init__',
                                                                                  'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D.forward': ( 'models/layers.html#resblock2d.forward',
                                                                                 'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D_Conditional': ( 'models/layers.html#resblock2d_conditional',
                                                                                     'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D_Conditional.__init__': ( 'models/layers.html#resblock2d_conditional.__init__',
                                                                                              'genQC/models/layers.py'),
                                     'genQC.models.layers.ResBlock2D_Conditional.forward': ( 'models/layers.html#resblock2d_conditional.forward',
                                                                                             'genQC/models/layers.py'),
                                     'genQC.models.layers.ResDownBlock2D': ('models/layers.html#resdownblock2d', 'genQC/models/layers.py'),
                                     'genQC.models.layers.ResDownBlock2D.__init__': ( 'models/layers.html#resdownblock2d.__init__',
                                                                                      'genQC/models/layers.py'),
                                     'genQC.models.layers.ResDownBlock2D.forward': ( 'models/layers.html#resdownblock2d.forward',
                                                                                     'genQC/models/layers.py'),
                                     'genQC.models.layers.ResUpBlock2D': ('models/layers.html#resupblock2d', 'genQC/models/layers.py'),
                                     'genQC.models.layers.ResUpBlock2D.__init__': ( 'models/layers.html#resupblock2d.__init__',
                                                                                    'genQC/models/layers.py'),
                                     'genQC.models.layers.ResUpBlock2D.forward': ( 'models/layers.html#resupblock2d.forward',
                                                                                   'genQC/models/layers.py'),
                                     'genQC.models.layers.TimeEmbedding': ('models/layers.html#timeembedding', 'genQC/models/layers.py'),
                                     'genQC.models.layers.TimeEmbedding.__init__': ( 'models/layers.html#timeembedding.__init__',
                                                                                     'genQC/models/layers.py'),
                                     'genQC.models.layers.TimeEmbedding.forward': ( 'models/layers.html#timeembedding.forward',
                                                                                    'genQC/models/layers.py'),
                                     'genQC.models.layers.UpBlock2D': ('models/layers.html#upblock2d', 'genQC/models/layers.py'),
                                     'genQC.models.layers.UpBlock2D.__init__': ( 'models/layers.html#upblock2d.__init__',
                                                                                 'genQC/models/layers.py'),
                                     'genQC.models.layers.UpBlock2D.forward': ( 'models/layers.html#upblock2d.forward',
                                                                                'genQC/models/layers.py')},
            'genQC.models.transformers': { 'genQC.models.transformers.BasisCrossAttnBlock': ( 'models/transformers.html#basiscrossattnblock',
                                                                                              'genQC/models/transformers.py'),
                                           'genQC.models.transformers.BasisCrossAttnBlock.__init__': ( 'models/transformers.html#basiscrossattnblock.__init__',
                                                                                                       'genQC/models/transformers.py'),
                                           'genQC.models.transformers.BasisCrossAttnBlock.forward': ( 'models/transformers.html#basiscrossattnblock.forward',
                                                                                                      'genQC/models/transformers.py'),
                                           'genQC.models.transformers.BasisSelfAttnBlock': ( 'models/transformers.html#basisselfattnblock',
                                                                                             'genQC/models/transformers.py'),
                                           'genQC.models.transformers.BasisSelfAttnBlock.__init__': ( 'models/transformers.html#basisselfattnblock.__init__',
                                                                                                      'genQC/models/transformers.py'),
                                           'genQC.models.transformers.BasisSelfAttnBlock.forward': ( 'models/transformers.html#basisselfattnblock.forward',
                                                                                                     'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformer': ( 'models/transformers.html#spatialtransformer',
                                                                                             'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformer.__init__': ( 'models/transformers.html#spatialtransformer.__init__',
                                                                                                      'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformer.forward': ( 'models/transformers.html#spatialtransformer.forward',
                                                                                                     'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformerSelfAttn': ( 'models/transformers.html#spatialtransformerselfattn',
                                                                                                     'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformerSelfAttn.__init__': ( 'models/transformers.html#spatialtransformerselfattn.__init__',
                                                                                                              'genQC/models/transformers.py'),
                                           'genQC.models.transformers.SpatialTransformerSelfAttn.forward': ( 'models/transformers.html#spatialtransformerselfattn.forward',
                                                                                                             'genQC/models/transformers.py')},
            'genQC.models.unet_qc': { 'genQC.models.unet_qc.Decoder': ('models/unet_qc.html#decoder', 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.Decoder.__init__': ( 'models/unet_qc.html#decoder.__init__',
                                                                                 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.Decoder.forward': ( 'models/unet_qc.html#decoder.forward',
                                                                                'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.Encoder': ('models/unet_qc.html#encoder', 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.Encoder.__init__': ( 'models/unet_qc.html#encoder.__init__',
                                                                                 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.Encoder.forward': ( 'models/unet_qc.html#encoder.forward',
                                                                                'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Compilation_UNet': ( 'models/unet_qc.html#qc_compilation_unet',
                                                                                    'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Compilation_UNet.__init__': ( 'models/unet_qc.html#qc_compilation_unet.__init__',
                                                                                             'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Compilation_UNet.forward': ( 'models/unet_qc.html#qc_compilation_unet.forward',
                                                                                            'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Compilation_UNet_config': ( 'models/unet_qc.html#qc_compilation_unet_config',
                                                                                           'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet': ('models/unet_qc.html#qc_cond_unet', 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet.__init__': ( 'models/unet_qc.html#qc_cond_unet.__init__',
                                                                                      'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet._init_weights': ( 'models/unet_qc.html#qc_cond_unet._init_weights',
                                                                                           'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet.embedd_clrs': ( 'models/unet_qc.html#qc_cond_unet.embedd_clrs',
                                                                                         'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet.forward': ( 'models/unet_qc.html#qc_cond_unet.forward',
                                                                                     'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet.invert_clr': ( 'models/unet_qc.html#qc_cond_unet.invert_clr',
                                                                                        'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.QC_Cond_UNet_config': ( 'models/unet_qc.html#qc_cond_unet_config',
                                                                                    'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.UNet_block': ('models/unet_qc.html#unet_block', 'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.UNet_block.__init__': ( 'models/unet_qc.html#unet_block.__init__',
                                                                                    'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.UNet_block._init_weights': ( 'models/unet_qc.html#unet_block._init_weights',
                                                                                         'genQC/models/unet_qc.py'),
                                      'genQC.models.unet_qc.UNet_block.forward': ( 'models/unet_qc.html#unet_block.forward',
                                                                                   'genQC/models/unet_qc.py')},
            'genQC.models.unitary_encoder': { 'genQC.models.unitary_encoder.Unitary_encoder': ( 'models/unitary_encoder.html#unitary_encoder',
                                                                                                'genQC/models/unitary_encoder.py'),
                                              'genQC.models.unitary_encoder.Unitary_encoder.__init__': ( 'models/unitary_encoder.html#unitary_encoder.__init__',
                                                                                                         'genQC/models/unitary_encoder.py'),
                                              'genQC.models.unitary_encoder.Unitary_encoder._init_weights': ( 'models/unitary_encoder.html#unitary_encoder._init_weights',
                                                                                                              'genQC/models/unitary_encoder.py'),
                                              'genQC.models.unitary_encoder.Unitary_encoder.forward': ( 'models/unitary_encoder.html#unitary_encoder.forward',
                                                                                                        'genQC/models/unitary_encoder.py'),
                                              'genQC.models.unitary_encoder.Unitary_encoder_config': ( 'models/unitary_encoder.html#unitary_encoder_config',
                                                                                                       'genQC/models/unitary_encoder.py')},
            'genQC.pipeline.diffusion_pipeline': { 'genQC.pipeline.diffusion_pipeline.DiffusionPipeline': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline',
                                                                                                            'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.CFG': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.cfg',
                                                                                                                'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.__call__': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.__call__',
                                                                                                                     'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.__init__': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.__init__',
                                                                                                                     'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.denoising': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.denoising',
                                                                                                                      'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.denoising_step': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.denoising_step',
                                                                                                                           'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.from_config_file': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.from_config_file',
                                                                                                                             'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.from_pretrained': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.from_pretrained',
                                                                                                                            'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.get_guidance_condition': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.get_guidance_condition',
                                                                                                                                   'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.latent_filling': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.latent_filling',
                                                                                                                           'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.params_config': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.params_config',
                                                                                                                          'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.prepare_c_emb': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.prepare_c_emb',
                                                                                                                          'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.store_pipeline': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.store_pipeline',
                                                                                                                           'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.train_on_epoch': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.train_on_epoch',
                                                                                                                           'genQC/pipeline/diffusion_pipeline.py'),
                                                   'genQC.pipeline.diffusion_pipeline.DiffusionPipeline.train_step': ( 'pipeline/diffusion_pipeline.html#diffusionpipeline.train_step',
                                                                                                                       'genQC/pipeline/diffusion_pipeline.py')},
            'genQC.pipeline.diffusion_pipeline_special': { 'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation',
                                                                                                                                        'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation.__call__': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation.__call__',
                                                                                                                                                 'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation.denoising': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation.denoising',
                                                                                                                                                  'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation.denoising_step': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation.denoising_step',
                                                                                                                                                       'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation.get_guidance_U': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation.get_guidance_u',
                                                                                                                                                       'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_Compilation.train_step': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_compilation.train_step',
                                                                                                                                                   'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_attnPadded': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_attnpadded',
                                                                                                                                       'genQC/pipeline/diffusion_pipeline_special.py'),
                                                           'genQC.pipeline.diffusion_pipeline_special.DiffusionPipeline_attnPadded.train_step': ( 'pipeline/diffusion_pipeline_special.html#diffusionpipeline_attnpadded.train_step',
                                                                                                                                                  'genQC/pipeline/diffusion_pipeline_special.py')},
            'genQC.pipeline.pipeline': { 'genQC.pipeline.pipeline.Pipeline': ( 'pipeline/pipeline.html#pipeline',
                                                                               'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.__call__': ( 'pipeline/pipeline.html#pipeline.__call__',
                                                                                        'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.__init__': ( 'pipeline/pipeline.html#pipeline.__init__',
                                                                                        'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline._reset_opt': ( 'pipeline/pipeline.html#pipeline._reset_opt',
                                                                                          'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline._set_opt_param': ( 'pipeline/pipeline.html#pipeline._set_opt_param',
                                                                                              'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.compile': ( 'pipeline/pipeline.html#pipeline.compile',
                                                                                       'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.end_batch_metrics': ( 'pipeline/pipeline.html#pipeline.end_batch_metrics',
                                                                                                 'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.end_epoch_metrics': ( 'pipeline/pipeline.html#pipeline.end_epoch_metrics',
                                                                                                 'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.end_progress_bar_iteration': ( 'pipeline/pipeline.html#pipeline.end_progress_bar_iteration',
                                                                                                          'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.fit': ( 'pipeline/pipeline.html#pipeline.fit',
                                                                                   'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.fit_summary': ( 'pipeline/pipeline.html#pipeline.fit_summary',
                                                                                           'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.progress_bar': ( 'pipeline/pipeline.html#pipeline.progress_bar',
                                                                                            'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.summary': ( 'pipeline/pipeline.html#pipeline.summary',
                                                                                       'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.train_on_batch': ( 'pipeline/pipeline.html#pipeline.train_on_batch',
                                                                                              'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.train_on_epoch': ( 'pipeline/pipeline.html#pipeline.train_on_epoch',
                                                                                              'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline.train_step': ( 'pipeline/pipeline.html#pipeline.train_step',
                                                                                          'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline_IO': ( 'pipeline/pipeline.html#pipeline_io',
                                                                                  'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline_IO.from_config_file': ( 'pipeline/pipeline.html#pipeline_io.from_config_file',
                                                                                                   'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline_IO.get_config': ( 'pipeline/pipeline.html#pipeline_io.get_config',
                                                                                             'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline_IO.params_config': ( 'pipeline/pipeline.html#pipeline_io.params_config',
                                                                                                'genQC/pipeline/pipeline.py'),
                                         'genQC.pipeline.pipeline.Pipeline_IO.store_pipeline': ( 'pipeline/pipeline.html#pipeline_io.store_pipeline',
                                                                                                 'genQC/pipeline/pipeline.py')},
            'genQC.platform.qcircuit_dataset_construction': { 'genQC.platform.qcircuit_dataset_construction.decode_circuit': ( 'platform/qcircuit_dataset_construction.html#decode_circuit',
                                                                                                                               'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.encode_circuit': ( 'platform/qcircuit_dataset_construction.html#encode_circuit',
                                                                                                                               'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.gen_compilation_rndGates_dataset': ( 'platform/qcircuit_dataset_construction.html#gen_compilation_rndgates_dataset',
                                                                                                                                                 'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.gen_qc_dataset': ( 'platform/qcircuit_dataset_construction.html#gen_qc_dataset',
                                                                                                                               'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.get_rnd_encoded_circuit': ( 'platform/qcircuit_dataset_construction.html#get_rnd_encoded_circuit',
                                                                                                                                        'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.get_rnd_encoded_circuits': ( 'platform/qcircuit_dataset_construction.html#get_rnd_encoded_circuits',
                                                                                                                                         'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.get_specific_rnd_srv_circuit': ( 'platform/qcircuit_dataset_construction.html#get_specific_rnd_srv_circuit',
                                                                                                                                             'genQC/platform/qcircuit_dataset_construction.py'),
                                                              'genQC.platform.qcircuit_dataset_construction.get_target_control_qubits': ( 'platform/qcircuit_dataset_construction.html#get_target_control_qubits',
                                                                                                                                          'genQC/platform/qcircuit_dataset_construction.py')},
            'genQC.platform.qcircuit_evaluation': { 'genQC.platform.qcircuit_evaluation.extract_gate_number': ( 'platform/qcircuit_evaluation.html#extract_gate_number',
                                                                                                                'genQC/platform/qcircuit_evaluation.py'),
                                                    'genQC.platform.qcircuit_evaluation.get_gate_stat_from_circuits': ( 'platform/qcircuit_evaluation.html#get_gate_stat_from_circuits',
                                                                                                                        'genQC/platform/qcircuit_evaluation.py'),
                                                    'genQC.platform.qcircuit_evaluation.get_gate_stat_from_tensors': ( 'platform/qcircuit_evaluation.html#get_gate_stat_from_tensors',
                                                                                                                       'genQC/platform/qcircuit_evaluation.py'),
                                                    'genQC.platform.qcircuit_evaluation.sort_into_bins': ( 'platform/qcircuit_evaluation.html#sort_into_bins',
                                                                                                           'genQC/platform/qcircuit_evaluation.py')},
            'genQC.platform.qcircuit_metrics': { 'genQC.platform.qcircuit_metrics.Unitary_FrobeniusNorm': ( 'platform/qcircuit_metrics.html#unitary_frobeniusnorm',
                                                                                                            'genQC/platform/qcircuit_metrics.py'),
                                                 'genQC.platform.qcircuit_metrics.Unitary_FrobeniusNorm.distance': ( 'platform/qcircuit_metrics.html#unitary_frobeniusnorm.distance',
                                                                                                                     'genQC/platform/qcircuit_metrics.py'),
                                                 'genQC.platform.qcircuit_metrics.Unitary_FrobeniusNorm.name': ( 'platform/qcircuit_metrics.html#unitary_frobeniusnorm.name',
                                                                                                                 'genQC/platform/qcircuit_metrics.py')},
            'genQC.platform.qcircuit_util': { 'genQC.platform.qcircuit_util.get_element_matching_indices': ( 'platform/qcircuit_util.html#get_element_matching_indices',
                                                                                                             'genQC/platform/qcircuit_util.py'),
                                              'genQC.platform.qcircuit_util.get_entanglement_bins': ( 'platform/qcircuit_util.html#get_entanglement_bins',
                                                                                                      'genQC/platform/qcircuit_util.py')},
            'genQC.platform.simulation.qcircuit_sim': { 'genQC.platform.simulation.qcircuit_sim.gate_pool_to_gate_classes': ( 'platform/simulation/qcircuit_sim.html#gate_pool_to_gate_classes',
                                                                                                                              'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.get_number_of_gate_params': ( 'platform/simulation/qcircuit_sim.html#get_number_of_gate_params',
                                                                                                                              'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.instruction_name_to_qiskit_gate': ( 'platform/simulation/qcircuit_sim.html#instruction_name_to_qiskit_gate',
                                                                                                                                    'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.optimize_circuit': ( 'platform/simulation/qcircuit_sim.html#optimize_circuit',
                                                                                                                     'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.plot_svr_stat': ( 'platform/simulation/qcircuit_sim.html#plot_svr_stat',
                                                                                                                  'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.rnd_circuit': ( 'platform/simulation/qcircuit_sim.html#rnd_circuit',
                                                                                                                'genQC/platform/simulation/qcircuit_sim.py'),
                                                        'genQC.platform.simulation.qcircuit_sim.schmidt_rank_vector': ( 'platform/simulation/qcircuit_sim.html#schmidt_rank_vector',
                                                                                                                        'genQC/platform/simulation/qcircuit_sim.py')},
            'genQC.printing': { 'genQC.printing.display_colums': ('printing.html#display_colums', 'genQC/printing.py'),
                                'genQC.printing.ndarray_to_latex': ('printing.html#ndarray_to_latex', 'genQC/printing.py'),
                                'genQC.printing.print_markdown': ('printing.html#print_markdown', 'genQC/printing.py'),
                                'genQC.printing.print_table': ('printing.html#print_table', 'genQC/printing.py'),
                                'genQC.printing.tensor_to_latex': ('printing.html#tensor_to_latex', 'genQC/printing.py')},
            'genQC.scheduler.scheduler': { 'genQC.scheduler.scheduler.Scheduler': ( 'scheduler/scheduler.html#scheduler',
                                                                                    'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.__init__': ( 'scheduler/scheduler.html#scheduler.__init__',
                                                                                             'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.add_noise': ( 'scheduler/scheduler.html#scheduler.add_noise',
                                                                                              'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.get_config': ( 'scheduler/scheduler.html#scheduler.get_config',
                                                                                               'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.params_config': ( 'scheduler/scheduler.html#scheduler.params_config',
                                                                                                  'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.set_timesteps': ( 'scheduler/scheduler.html#scheduler.set_timesteps',
                                                                                                  'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.step': ( 'scheduler/scheduler.html#scheduler.step',
                                                                                         'genQC/scheduler/scheduler.py'),
                                           'genQC.scheduler.scheduler.Scheduler.unsqueeze_vector_to_shape': ( 'scheduler/scheduler.html#scheduler.unsqueeze_vector_to_shape',
                                                                                                              'genQC/scheduler/scheduler.py')},
            'genQC.scheduler.scheduler_ddim': { 'genQC.scheduler.scheduler_ddim.DDIMScheduler': ( 'scheduler/scheduler_ddim.html#ddimscheduler',
                                                                                                  'genQC/scheduler/scheduler_ddim.py'),
                                                'genQC.scheduler.scheduler_ddim.DDIMScheduler.__init__': ( 'scheduler/scheduler_ddim.html#ddimscheduler.__init__',
                                                                                                           'genQC/scheduler/scheduler_ddim.py'),
                                                'genQC.scheduler.scheduler_ddim.DDIMScheduler.params_config': ( 'scheduler/scheduler_ddim.html#ddimscheduler.params_config',
                                                                                                                'genQC/scheduler/scheduler_ddim.py'),
                                                'genQC.scheduler.scheduler_ddim.DDIMScheduler.set_timesteps': ( 'scheduler/scheduler_ddim.html#ddimscheduler.set_timesteps',
                                                                                                                'genQC/scheduler/scheduler_ddim.py'),
                                                'genQC.scheduler.scheduler_ddim.DDIMScheduler.step': ( 'scheduler/scheduler_ddim.html#ddimscheduler.step',
                                                                                                       'genQC/scheduler/scheduler_ddim.py'),
                                                'genQC.scheduler.scheduler_ddim.DDIMSchedulerOutput': ( 'scheduler/scheduler_ddim.html#ddimscheduleroutput',
                                                                                                        'genQC/scheduler/scheduler_ddim.py')},
            'genQC.scheduler.scheduler_ddpm': { 'genQC.scheduler.scheduler_ddpm.DDPMScheduler': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler',
                                                                                                  'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.__init__': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.__init__',
                                                                                                           'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.add_noise': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.add_noise',
                                                                                                            'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.add_noise_LEdit': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.add_noise_ledit',
                                                                                                                  'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.params_config': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.params_config',
                                                                                                                'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.set_timesteps': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.set_timesteps',
                                                                                                                'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.step': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.step',
                                                                                                       'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMScheduler.to_device': ( 'scheduler/scheduler_ddpm.html#ddpmscheduler.to_device',
                                                                                                            'genQC/scheduler/scheduler_ddpm.py'),
                                                'genQC.scheduler.scheduler_ddpm.DDPMSchedulerOutput': ( 'scheduler/scheduler_ddpm.html#ddpmscheduleroutput',
                                                                                                        'genQC/scheduler/scheduler_ddpm.py')},
            'genQC.util': { 'genQC.util.DataLoaders': ('util.html#dataloaders', 'genQC/util.py'),
                            'genQC.util.DataLoaders.__init__': ('util.html#dataloaders.__init__', 'genQC/util.py'),
                            'genQC.util.MemoryCleaner': ('util.html#memorycleaner', 'genQC/util.py'),
                            'genQC.util.MemoryCleaner._clean_ipython_hist': ( 'util.html#memorycleaner._clean_ipython_hist',
                                                                              'genQC/util.py'),
                            'genQC.util.MemoryCleaner._clean_tb': ('util.html#memorycleaner._clean_tb', 'genQC/util.py'),
                            'genQC.util.MemoryCleaner.purge_mem': ('util.html#memorycleaner.purge_mem', 'genQC/util.py'),
                            'genQC.util.infer_torch_device': ('util.html#infer_torch_device', 'genQC/util.py'),
                            'genQC.util.latents_to_pil': ('util.html#latents_to_pil', 'genQC/util.py'),
                            'genQC.util.normalize_tensor': ('util.html#normalize_tensor', 'genQC/util.py'),
                            'genQC.util.number_of_paramters': ('util.html#number_of_paramters', 'genQC/util.py'),
                            'genQC.util.plot_image_grid': ('util.html#plot_image_grid', 'genQC/util.py'),
                            'genQC.util.savePdf': ('util.html#savepdf', 'genQC/util.py'),
                            'genQC.util.savePng': ('util.html#savepng', 'genQC/util.py'),
                            'genQC.util.saveSvg': ('util.html#savesvg', 'genQC/util.py'),
                            'genQC.util.scale_tensor': ('util.html#scale_tensor', 'genQC/util.py'),
                            'genQC.util.virtual': ('util.html#virtual', 'genQC/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/inference/infer_misc.ipynb.

# %% auto 0
__all__ = ['get_rnd_gatepool_subset', 'convert_tensors_to_circuits']

# %% ../../src/inference/infer_misc.ipynb 2
from ..imports import *
from ..platform.qcircuit_dataset_construction import decode_circuit, gate_pool_to_gate_classes
from ..platform.simulation.qcircuit_sim import instruction_name_to_qiskit_gate

# %% ../../src/inference/infer_misc.ipynb 4
def get_rnd_gatepool_subset(gate_pool, min_sub_gate_pool_cnt=2):
    rng = np.random.default_rng()
    
    n       = len(gate_pool) + 1
    c_range = np.arange(n-1)
    
    sub_gate_pool_cnt = rng.integers(min_sub_gate_pool_cnt, n)       
    sub_gate_pool_ind = rng.choice(c_range, size=sub_gate_pool_cnt, replace=False)         
    sub_gate_pool     = [gate_pool[ind] for ind in sub_gate_pool_ind]  # pick random subeset of gates
        
    return sub_gate_pool

# %% ../../src/inference/infer_misc.ipynb 6
def convert_tensors_to_circuits(out_tensor, gate_pool, params_tensor=None, place_barrier=False): 
    if isinstance(gate_pool[0], str):
        gate_pool = [instruction_name_to_qiskit_gate(gate) for gate in gate_pool]
    
    error_cnt = 0
    qc_list   = []

    if not exists(params_tensor):
        params_tensor = [None]*out_tensor.shape[0]
    
    #TODO: para this loop
   
    for i,(enc_tensor,p) in enumerate(zip(out_tensor, params_tensor)):   
        try:
            qc = decode_circuit(enc_tensor=enc_tensor, gate_pool=gate_pool, place_barrier=place_barrier, params_tensor=p)

        except Exception as e:
            error_cnt += 1
            # print(e)
            continue
                 
        qc_list.append(qc)  
  
    return qc_list, error_cnt



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/dataset/cached_qc_dataset.ipynb.

# %% auto 0
__all__ = ['Cached_OpenClip_Dataset']

# %% ../../src/dataset/cached_qc_dataset.ipynb 3
from ..imports import *
from .qc_dataset import Qc_Config_Dataset
from .config_dataset import Config_Dataset
from ..config_loader import *

# %% ../../src/dataset/cached_qc_dataset.ipynb 4
class Cached_OpenClip_Dataset(Qc_Config_Dataset):
    """Adds `.caching` to the `Quantum circuit dataset` class."""
    
    def x_y_preprocess(self, balance_max, max_samples=None):
        x_proc, y_proc, *z = super().x_y_preprocess(balance_max=balance_max, max_samples=max_samples)        
        y_proc = self.caching(y_proc)
        return x_proc, y_proc, *z
    
    def caching(self, y_proc, y_on_cpu=False):
        print("[INFO]: Generate cache: converting tensors to str and tokenize")   
        
        print(" - to str list")  
        if isinstance(y_proc, (torch.Tensor, torch.IntTensor, torch.FloatTensor, torch.LongTensor)):         
            y_str = [str(i) for i in y_proc.cpu().tolist()]
        elif isinstance(y_proc, list): 
            y_str = []
            for iy in y_proc:                
                if isinstance(iy, np.ndarray): y_str += [str(i) for i in iy.tolist()]        # list of numpy arrays
                else:                          y_str += [str(i) for i in iy.cpu().tolist()]  # list of tensors
        elif isinstance(y_proc, np.ndarray):
            y_str = [str(i) for i in y_proc.tolist()]
            
        else: raise NotImplementedError()
                            
        print(" - tokenize_and_push_to_device")  
        y_tok = self.text_encoder.tokenize_and_push_to_device(y_str, to_device= not y_on_cpu)
        if y_on_cpu: y_tok = y_tok.cpu()
        
        
        #now for using cache we need the uniques and the corresponding indices of the uniques
        y_uniques, y_ptrs  = torch.unique(torch.cat([self.text_encoder.empty_token.to(y_tok.device), y_tok]), dim=0, return_inverse=True)
    
        cached_empty_token_index = y_ptrs[0]  #store what index the empty token has   
        y_ptrs                   = y_ptrs[1:] #remove the cat empty token
      
        #use cache
        print(" - generate_cache")  
        self.text_encoder.generate_cache(tokens=y_uniques, cached_empty_token_index=cached_empty_token_index, y_on_cpu=y_on_cpu)
      
        print("[INFO]: Generated cache")  
        return y_ptrs
    
    #-------------------------------------------
    
    def get_dataloaders(self, batch_size, text_encoder, p_valid=0.1, balance_max=None, max_samples=None):
        self.text_encoder = text_encoder    
        return super().get_dataloaders(batch_size, p_valid, balance_max, max_samples)
       
    #-------------------------------------------
        
    @staticmethod
    def from_config_file(config_path, device: torch.device, save_path: str=None):
        config = load_config(config_path)
        config["target"] = class_to_str(Cached_OpenClip_Dataset)               
        return Config_Dataset.from_config(config, device, save_path)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/dataset/mixed_cached_qc_dataset.ipynb.

# %% auto 0
__all__ = ['Mixed_Cached_OpenClip_Dataset_config', 'Mixed_Cached_OpenClip_Dataset']

# %% ../../src/dataset/mixed_cached_qc_dataset.ipynb 3
from ..imports import *
from .qc_dataset import Qc_Config_Dataset_config, Qc_Config_Dataset
from .config_dataset import Config_Dataset
from .cached_qc_dataset import Cached_OpenClip_Dataset
from ..config_loader import *
from .dataset_helper import *
from ..util import DataLoaders
import dataclasses

# %% ../../src/dataset/mixed_cached_qc_dataset.ipynb 4
@dataclass
class Mixed_Cached_OpenClip_Dataset_config(Qc_Config_Dataset_config):
    pad_constant: int
    collate_fn: str
    bucket_batch_size: int
    num_down_scales: int  # for flex pad attn mask

# %% ../../src/dataset/mixed_cached_qc_dataset.ipynb 5
class Mixed_Cached_OpenClip_Dataset(Cached_OpenClip_Dataset):  
    """Dataset that uses multiple cached dataset and combines them with padding, either i) Bucket or  ii) Max. Also provides a corresponding `collate_fn` for training."""
    
    req_params = [f.name for f in dataclasses.fields(Mixed_Cached_OpenClip_Dataset_config)]
     
    cut_multiple = 4  #needed for proper downscaling!
    
    @property
    def params_config(self):
        params_config = {}
        for p in self.req_params: params_config[p] = getattr(self, p)   
        params_config["gate_pool"] = [class_to_str(gate) for gate in params_config["gate_pool"]]
        params_config = Mixed_Cached_OpenClip_Dataset_config(**params_config)
        return params_config  
        
    #-----------------------------------
    # CAUSAL ATTENTION PADDING

    def flexPadAttn_padding_collate_fn(self, b):  
        """this function is called for training for every batch"""
        z_0 = max(x[2][0] for x in b)  # space
        z_1 = max(x[2][1] for x in b)  # time
        
        #round time to next multiple of 8 for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------
        # key_padding_mask ... [N, S]    -inf where we want no attention
        # we will create here [N, s, t] and then reshaping is easy
        # note this is key pad mask not directly attention mask! we need this for loss masking
        # Nb: add rnd to the padding, so we train with pad and on smaller systems
        
        #we need 3 different ones for the different unet layers        
        key_padding_mask = torch.zeros((len(b), z_0, z_1), device=self.device)  
              
        padd_rnds = torch.randint(low=0, high=2, size=(len(b),2), dtype=torch.int32)    #roll 50/50 if we allow padding
                 
        xs=[]
        ys=[]
        for i,((x,y,z), padd_rnd) in enumerate(zip(b, padd_rnds)):
        # for i,(x,y,z) in enumerate(b):
            x = x[:z_0, :z_1]  # cut down to max [bits, time] of batch
            
            #-------------------  
            space, time = z[0], z[1]
        
            if space < z_0 and padd_rnd[0]: space = torch.randint(low=space, high=z_0+1, size=(1,), dtype=torch.int32)             
            if time  < z_1 and padd_rnd[1]: time  = torch.randint(low=time , high=z_1+1, size=(1,), dtype=torch.int32)                  
            
            time = (torch.ceil(time / self.cut_multiple) * self.cut_multiple).to(torch.int32)
            
            key_padding_mask[i, space:,     :] = float('-inf') 
            key_padding_mask[i,      :, time:] = float('-inf')     
            
            #-------------------   
            
            xs.append(x)
            ys.append(y)
               
        key_padding_mask_list = [key_padding_mask]
        for j in range(1, self.num_down_scales):
            key_padding_mask_list.append(F.max_pool1d(key_padding_mask_list[j-1], kernel_size=2))   
                     
        xs=torch.stack(xs)
        ys=torch.stack(ys)  
        return xs, ys, key_padding_mask_list            
   
    def flexPadAttn_TimeOnly_padding_collate_fn(self, b):  
        """this function is called for training for every batch"""
        z_0 = max(x[2][0] for x in b)  # space
        z_1 = max(x[2][1] for x in b)  # time
        
        #round time to next multiple of 8 for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------
        # key_padding_mask ... [N, S]    -inf where we want no attention
        # we will create here [N, s, t] and then reshaping is easy
        # note this is key pad mask not directly attention mask! we need this for loss masking
        # Nb: add rnd to the padding, so we train with pad and on smaller systems
        
        #we need 3 different ones for the different unet layers        
        key_padding_mask = torch.zeros((len(b), z_0, z_1), device=self.device)  
              
        padd_rnds = torch.randint(low=0, high=2, size=(len(b)), dtype=torch.int32)    #roll 50/50 if we allow padding
                 
        xs=[]
        ys=[]
        for i,((x,y,z), padd_rnd) in enumerate(zip(b, padd_rnds)):
        # for i,(x,y,z) in enumerate(b):
            x = x[:z_0, :z_1]  # cut down to max [bits, time] of batch
            
            #-------------------  
            time = z[1]
               
            if time  < z_1 and padd_rnd: time  = torch.randint(low=time , high=z_1+1, size=(1,), dtype=torch.int32)                       
            time = (torch.ceil(time / self.cut_multiple) * self.cut_multiple).to(torch.int32)            
            key_padding_mask[i, :, time:] = float('-inf')     
            
            #-------------------   
            
            xs.append(x)
            ys.append(y)
               
        key_padding_mask_list = [key_padding_mask]
        for j in range(1, self.num_down_scales):
            key_padding_mask_list.append(F.max_pool1d(key_padding_mask_list[j-1], kernel_size=2))   
                     
        xs=torch.stack(xs)
        ys=torch.stack(ys)  
        return xs, ys, key_padding_mask_list            

    #-----------------------------------
    # BUCKET PADDING, all x,y are already passed as batch
        
    def cut_padding_Bucket_collate_fn(self, b):     
        """this function is called for training for every batch"""    
        
        b = b[0]
        
        x = b[0]
        y = b[1]
        z = b[2]
                
        #---------------
        
        z_0 = torch.max(z[:, 0]) # space
        z_1 = torch.max(z[:, 1]) # time
                   
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------      
        
        x = x[:, :z_0, :z_1]  # cut down to max [b, bits, time] of batch
        
        return x, y

    def cut_padding_Bucket_collate_fn_compilation(self, b):     
        """this function is called for training for every batch"""    
        
        b = b[0]
        
        x = b[0]
        y = b[1]                
        U = b[2]
        z = b[3]
        
        #---------------
        
        z_0 = torch.max(z[:, 0]) # space
        z_1 = torch.max(z[:, 1]) # time
                   
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------      
        
        x = x[:, :z_0, :z_1]  # cut down to max [b, bits, time] of batch
        
        bit_exp = 2**z_0
        U = U[:, :, :bit_exp, :bit_exp]   # [b, Re/Im, 2^n, 2^n]
               
        return x, y, U

    def cut_padding_Bucket_collate_fn_compilation_params(self, b):     
        """this function is called for training for every batch, order in b is store dict"""    
        
        b = b[0] # {'x': 'tensor', 'y': 'numpy', 'params': 'tensor', 'U': 'tensor', 'z': 'tensor'}
        
        x = b[0]
        y = b[1]  
        p = b[2]
        U = b[3]
        z = b[4]
        
        #---------------
        
        z_0 = torch.max(z[:, 0]) # space
        z_1 = torch.max(z[:, 1]) # time
                   
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------      
        
        x = x[:, :z_0, :z_1]  # cut down to max [b, bits, time] of batch

        p = p[:, :, :z_1]
        
        bit_exp = 2**z_0
        U = U[:, :, :bit_exp, :bit_exp]   # [b, Re/Im, 2^n, 2^n]
               
        return x, y, p, U
    
    #-----------------------------------
    # MAX PADDING, x are passes as sampled list (batch), std collate them
    
    def cut_padding_collate_fn(self, b):     
        """this function is called for training for every batch"""    
        z_0 = max(x[2][0] for x in b)  # space
        z_1 = max(x[2][1] for x in b)  # time
        
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)
              
        #---------------      

        x_sample = b[0][0]
        xs       = torch.zeros((len(b), z_0, z_1), dtype=x_sample.dtype, device=x_sample.device)
        
        # xs=[]
        ys=[]
        for i,(x,y,z) in enumerate(b):
            #x = x[:z_0, :z_1]  # cut down to max [bits, time] of batch
            xs[i] = x[:z_0, :z_1]
            
            #xs.append(x)
            ys.append(y)
                
        #xs=torch.stack(xs)
        ys=torch.stack(ys)  
         
        return xs, ys   

    def cut_padding_collate_fn_compilation(self, b):
        """this function is called for training for every batch"""    
        z_0 = max(x[3][0] for x in b)  # space
        z_1 = max(x[3][1] for x in b)  # time
        
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)

        bit_exp = 2**z_0
        
        #---------------      

        x_sample = b[0][0]
        xs       = torch.zeros((len(b), z_0, z_1), dtype=x_sample.dtype, device=x_sample.device)

        y_sample = b[0][1]
        ys       = torch.zeros((len(b), *y_sample.shape), dtype=y_sample.dtype, device=y_sample.device)

        U_sample = b[0][2]
        Us       = torch.zeros((len(b), 2, bit_exp, bit_exp), dtype=U_sample.dtype, device=U_sample.device)
        
        for i,(x,y,U,z) in enumerate(b):
            xs[i] = x[:z_0, :z_1]
            ys[i] = y
            Us[i] = U[:, :bit_exp, :bit_exp]
         
        return xs, ys, Us   

    def cut_padding_collate_fn_compilation_params(self, b):
        """this function is called for training for every batch, order in b is store dict"""    
        # {'x': 'tensor', 'y': 'numpy', 'params': 'tensor', 'U': 'tensor', 'z': 'tensor'}
        
        z_0 = max(x[4][0] for x in b)  # space
        z_1 = max(x[4][1] for x in b)  # time
        
        #round time to next multiple of cut_multiple for conv layers!
        z_1 = (torch.ceil(z_1 / self.cut_multiple) * self.cut_multiple).to(torch.int32)

        bit_exp = 2**z_0
        
        #---------------      

        x_sample = b[0][0]
        xs       = torch.zeros((len(b), z_0, z_1), dtype=x_sample.dtype, device=x_sample.device)

        y_sample = b[0][1]
        ys       = torch.zeros((len(b), *y_sample.shape), dtype=y_sample.dtype, device=y_sample.device)

        p_sample = b[0][2]
        ps       = torch.zeros((len(b), p_sample.shape[-2], z_1), dtype=p_sample.dtype, device=p_sample.device)
        
        U_sample = b[0][3]
        Us       = torch.zeros((len(b), 2, bit_exp, bit_exp), dtype=U_sample.dtype, device=U_sample.device)
        
        for i,(x,y,p,U,z) in enumerate(b):
            xs[i] = x[:z_0, :z_1]
            ys[i] = y
            ps[i] = p[:, :z_1]
            Us[i] = U[:, :bit_exp, :bit_exp]
         
        return xs, ys, ps, Us   
    
    #-----------------------------------

    def get_dataloaders(self, batch_size, text_encoder, p_valid=0.1, y_on_cpu=False):
        self.text_encoder = text_encoder
        
        excepts = []
        if y_on_cpu: excepts.append("y")
        if self.params_config.dataset_to_gpu: self.to("cuda", excepts=excepts)
        
        x_proc, y_proc, *z_proc = Qc_Config_Dataset.x_y_preprocess(self, balance_max=None, shuffle=False)    # ... z_proc is `'z' and all other 'c'
        
        if self.bucket_batch_size <= 0:        
            y_proc = self.caching(y_proc, y_on_cpu=y_on_cpu)
                      
        else:                    
            y_proc = self.caching([yi.reshape((-1)) for yi in y_proc], y_on_cpu=y_on_cpu)
            y_proc = y_proc.reshape((-1, self.bucket_batch_size))
        
        x_proc, y_proc, *z_proc              = shuffle_tensor_dataset(x_proc, y_proc, *z_proc) #only possible after str y is cached as tensor
        x, x_valid, y, y_valid, (z, z_valid) = self.valid_split(x_proc, y_proc, *z_proc, p_valid=p_valid)
                             
        ds       = TensorDataset(x, y, *z)
        ds_valid = TensorDataset(x_valid, y_valid, *z_valid)

        collate_fn = getattr(self, self.collate_fn)
        
        if self.params_config.dataset_to_gpu: 
            train_loader = DataLoader(dataset=ds      , batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
            valid_loader = DataLoader(dataset=ds_valid, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

        else:              
            train_loader = DataLoader(dataset=ds      , batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=12, collate_fn=collate_fn)
            valid_loader = DataLoader(dataset=ds_valid, batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=12, collate_fn=collate_fn)

        self.dataloaders = DataLoaders(train_loader, valid_loader)        
        return self.dataloaders
    
    #-----------------------------------
    
    @staticmethod
    def from_datasets(datasets: list[Qc_Config_Dataset], balance_maxes: list, pad_constant, device: torch.device=torch.device("cpu"), bucket_batch_size=None, max_samples=None, **parameters):
        assert pad_constant != 0, "can NOT be 0! and not any other gate!"
        
        xs = []
        ys = []
        zs = []
        cs = []
        
        cut_multiple = Mixed_Cached_OpenClip_Dataset.cut_multiple
        
        max_qubits = max(dataset.params_config.num_of_qubits for dataset in datasets)
        max_gates  = max(dataset.params_config.max_gates     for dataset in datasets)
        max_gates  = int(np.ceil(max_gates /cut_multiple) * cut_multiple)
        
        parameters["num_of_qubits"]     = max_qubits
        parameters["max_gates"]         = max_gates
        parameters["random_samples"]    = sum([dataset.params_config.random_samples for dataset in datasets])
        parameters["min_gates"]         = min([dataset.params_config.min_gates      for dataset in datasets])
        parameters["comment"]           = f"Generated with 'from_datasets' with {len(datasets)} datasets. Qubits: {[dataset.params_config.num_of_qubits for dataset in datasets]}."
        parameters["pad_constant"]      = pad_constant
        parameters["bucket_batch_size"] = bucket_batch_size
         
        parameters["store_dict"] = {}
        for dataset in datasets:
            parameters["store_dict"] |= dataset.params_config.store_dict   #needs python 3.9 for union of dict  
        parameters["store_dict"]["z"]   = "tensor" #add special item

        if isinstance(max_samples, int):
            max_samples = [max_samples] * len(datasets)
        else:
            assert isinstance(max_samples, (list, np.ndarray))
            max_samples = np.array(max_samples, dtype=int)

        if isinstance(balance_maxes, int):
            balance_maxes = [balance_maxes] * len(datasets)
        else:
            assert isinstance(balance_maxes, (list, np.ndarray))
            balance_maxes = np.array(balance_maxes, dtype=int)
        
        for i, (dataset, balance_max) in tqdm(enumerate(zip(datasets,balance_maxes)), total=len(datasets)):
            # do x_y_preprocess now, we can't balance all together with mixed conditions
    
            dataset = dataset.to(device)
            
            x, y, *c = dataset.x_y_preprocess(balance_max=balance_max, max_samples=max_samples[i], shuffle=True)            
            x = x.to(device)    # [b, s, t]   
            
            print(f" - dataset size after balancing {x.shape[0]}")

            #-------
            # store original size
            z = torch.zeros((x.shape[0], 2), device=device, dtype=torch.int32)
            z[:, 0] = max(dataset.params_config.num_of_qubits, 1)
            
            red_x   = torch.sum(x.abs(), dim=1)          # [b, t]   .. collaps the zeros to get circuit length
            z[:, 1] = torch.count_nonzero(red_x, dim=1)  # [b]         
            z[z[:, 1]==0, 1] = 1 # make sure we don*t have 0, so we cheat and set it to 1 (there's only 1 unique zero gate circuit anyways). Needed for padding attn mask                 
            
            for i in range(x.shape[0]):
                x[i, z[i, 0]:,        :] = pad_constant
                x[i,        :, z[i, 1]:] = pad_constant
                
            z[:, 1] = (torch.ceil(z[:, 1] / cut_multiple) * cut_multiple).to(torch.int32) #for cut needs multiple

            #-------
            # now pad x, padding is defined from last dim forward!        
            pad = (0, max_gates-dataset.params_config.max_gates, 0, max_qubits-dataset.params_config.num_of_qubits) 
            x   = F.pad(x, pad, "constant", pad_constant)
                         
            # if c is missing something of the union we set it to a zero tensor
            for k,v in parameters["store_dict"].items(): 
                if k != "x" and k != "y" and k != "z":
                    
                    if k not in dataset.params_config.store_dict:
                        empty_tensor = torch.zeros((1,), device=device)
                        
                        if k == "U": #scetchy hardcoded for compilation
                            empty_tensor = torch.zeros((x.shape[0], 2, 1, 1), device=device) # unitary is [b, Re/Im, 2^n, 2^n]
                        
                        assert len(c) == 0
                        c.append(empty_tensor) #scetchy bcs if c is not empty we could break ordering!!!
                
            #combine datasets
            xs.append(x.cpu())  
            ys.append(y)
            zs.append(z) 
            cs.append([*c])

            dataset = dataset.to("cpu") #helps with gpu mem overflowing
        #-----------------

        has_U = "U" in parameters["store_dict"]
        has_p = "params" in parameters["store_dict"]
        
        if bucket_batch_size > 0:
            collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn.__name__
            if has_U: 
                collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn_compilation.__name__
                if has_p: 
                    collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_Bucket_collate_fn_compilation_params.__name__
        
        else:
            collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn.__name__   
            if has_U: 
                collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn_compilation.__name__
                if has_p: 
                    collate_fn_name = Mixed_Cached_OpenClip_Dataset.cut_padding_collate_fn_compilation_params.__name__

        parameters["collate_fn"] = collate_fn_name
        
        #-----------------
        if bucket_batch_size > 0:
            for i, (xi,yi,zi, ci) in enumerate(zip(xs, ys, zs, cs)):  #cut rest of batch        
                b_mult = int(np.floor(xi.shape[0] / bucket_batch_size) * bucket_batch_size)  
                
                xs[i] = xi[None, :b_mult].reshape((b_mult//bucket_batch_size, bucket_batch_size, *xi.shape[1:])) 
                zs[i] = zi[None, :b_mult].reshape((b_mult//bucket_batch_size, bucket_batch_size, *zi.shape[1:]))
                
                t = parameters["store_dict"]["y"]
                if v == "tensor" or v == "numpy": 
                    ys[i] = yi[None, :b_mult].reshape((b_mult//bucket_batch_size, bucket_batch_size, *yi.shape[1:]))    
                else: raise NotImplementedError("")
            
                #----
                #For U, etc
                add_ind = 0
                for k,v in parameters["store_dict"].items(): 
                    if k != "x" and k != "y" and k != "z":                             
                        if v == "tensor" or v == "numpy": 
                            cs[i][add_ind] = ci[add_ind][None, :b_mult].reshape((b_mult//bucket_batch_size, bucket_batch_size, *ci[add_ind].shape[1:]))   
                        else: raise NotImplementedError("")                      
                        add_ind += 1                      
                      
        x = torch.cat(xs)
        y = ys                 # torch.cat(ys) is wrong,  y is list of numpy or str!! not a tensor
        z = torch.cat(zs)
        c = cs
        
        #-----------------    
        
        mixed_Cached_OpenClip_Dataset = Mixed_Cached_OpenClip_Dataset(device, **parameters)        
        mixed_Cached_OpenClip_Dataset.x = x
        mixed_Cached_OpenClip_Dataset.y = y
        mixed_Cached_OpenClip_Dataset.z = z
        
        add_ind = 0
        for k,v in parameters["store_dict"].items(): 
            if k != "x" and k != "y" and k != "z":   
                                     
                if v == "tensor" and k == "U":    # hardcoded U padding !!
                                           
                    n = sum([ci[add_ind].shape[0] for ci in c])
                    if bucket_batch_size > 0: shape = (n, bucket_batch_size, 2, 2**max_qubits, 2**max_qubits)
                    else:                     shape = (n,                    2, 2**max_qubits, 2**max_qubits)
                            
                    # allocating zeros is better memory wise than torch.cat(ci_s) and F.pad(ci, pad, "constant", 0)
                    mem = np.prod(shape) * c[0][add_ind].element_size() * 1e-9
                    print(f"[INFO]: allocate memory for {k} {shape} on {c[0][add_ind].device} approx. {mem:.3f} GB")
                    ci_s = torch.zeros(shape, device=c[0][add_ind].device)                 
                  
                    run_i = 0
                    for i,ci in enumerate(c):
                        ci = ci[add_ind]                                              
                        if bucket_batch_size > 0:  ci_s[run_i:run_i+ci.shape[0], :, :, :ci.shape[-2], :ci.shape[-1]] = ci                          
                        else:                      ci_s[run_i:run_i+ci.shape[0],    :, :ci.shape[-2], :ci.shape[-1]] = ci                 
                        run_i += ci.shape[0]

                elif v == "tensor" and k == "params": # hardcoded paramter padding !!

                    max_params = max(ci[add_ind].shape[-2] for ci in c)
                    
                    n = sum(ci[add_ind].shape[0] for ci in c)
                    if bucket_batch_size > 0: shape = (n, bucket_batch_size, max_params, max_gates)
                    else:                     shape = (n,                    max_params, max_gates)

                    # allocating zeros is better memory wise than torch.cat(ci_s) and F.pad(ci, pad, "constant", 0)
                    mem = np.prod(shape) * c[0][add_ind].element_size() * 1e-9
                    print(f"[INFO]: allocate memory for {k} {shape} on {c[0][add_ind].device} approx. {mem:.3f} GB")
                    ci_s = torch.zeros(shape, device=c[0][add_ind].device)                 
                  
                    run_i = 0
                    for i,ci in enumerate(c):
                        ci = ci[add_ind]                                              
                        if bucket_batch_size > 0:  ci_s[run_i:run_i+ci.shape[0], :, :ci.shape[-2], :ci.shape[-1]] = ci                          
                        else:                      ci_s[run_i:run_i+ci.shape[0],    :ci.shape[-2], :ci.shape[-1]] = ci                 
                        run_i += ci.shape[0]
                
                elif v == "numpy": raise NotImplementedError("")   
                else:              raise NotImplementedError("")           
                
                setattr(mixed_Cached_OpenClip_Dataset, str(k), ci_s)
                add_ind += 1
        
        return mixed_Cached_OpenClip_Dataset
    
    #------------------------------------
    
    # def plot_example(self):     print("plot_example not implemented for Mixed_Cached_OpenClip_Dataset")
    # def plot_distribution(self): print("plot_distribution not implemented for Mixed_Cached_OpenClip_Dataset")
    
    @staticmethod
    def from_config_file(config_path, device: torch.device, save_path: str=None):
        config = load_config(config_path)
        config["target"] = class_to_str(Mixed_Cached_OpenClip_Dataset)               
        return Config_Dataset.from_config(config, device, save_path)
